"""
Contains caching class for Collisionally Induced Absorption files
"""

from .singleton import Singleton
from taurex.log import Logger


class CIACache(Singleton):
    """
    Implements a lazy load of collisionally induced absorpiton cross-sections
    Supports pickle files and HITRAN cia files. Functionally behaves the same
    as :class:`~taurex.cache.opacitycache.OpacityCache` except the keys are
    now cia pairs
    e.g:

    >>> CIACache()['H2-H2']
    <taurex.cia.picklecia.PickleCIA at 0x107a60be0>

    Pickle ``.db`` and HITRAN ``.cia`` files are supported and automatically
    loaded. with priority given to ``.db`` files



    """
    def init(self):
        self.cia_dict = {}
        self._cia_path = None
        self.log = Logger('CIACache')

    def set_cia_path(self, cia_path):
        """

        Sets the path to search for CIA files

        Parameters
        ----------
        cia_path : str or :obj:`list` of str
            Either a single path or a list of paths that contain CIA files

        """
        self._cia_path = cia_path

    def __getitem__(self, key):
        """
        For a CIA pair, load from the set path and return the
        relevant :class:`~taurex.cia.cia.CIA` object

        Parameter
        ---------
        key : str
            cia pair name

        Returns
        -------
        :class:`~taurex.cia.picklecia.PickleCIA` or :class:`~taurex.cia.hitrancia.HitranCIA`
            Desire CIA object, format depends on what is found
            in the path set by :func:`set_cia_path`

        Raise
        -----
        Exception
            If desired CIA pair could not be loaded

        """

        if key in self.cia_dict:
            return self.cia_dict[key]
        else:
            # Try a load of the cia
            self.load_cia(pair_filter=[key])
            # If we have it after a load then good job boys
            if key in self.cia_dict:
                return self.cia_dict[key]
            else:
                # Otherwise throw an error
                self.log.error('CIA for pair %s could not be loaded', key)
                self.log.error('It could not be found in the local dictionary '
                               ' %s', list(self.cia_dict.keys()))
                self.log.error('Or paths %s', self._cia_path)
                self.log.error('Try loading it manually/ putting it in a path')
                raise Exception('cia could notn be loaded')

    def add_cia(self, cia, pair_filter=None):
        """

        Adds a :class:`~taurex.cia.cia.CIA` object to the cache to then be
        used by Taurex 3

        Parameters
        ----------
        cia : :class:`~taurex.cia.cia.CIA`
            CIA object to add to the cache

        pair_filter : :obj:`list` of str , optional
            If provided, the cia object will only be included
            if its pairname is in the list. Mostly used by the
            :func:`__getitem__` for filtering

        """

        self.log.info('Loading cia %s into model', cia.pairName)
        if cia.pairName in self.cia_dict:
            self.log.error('cia with name %s already in '
                           'opactiy dictionary %s', cia.pairName,
                           self.cia_dict.keys())

            raise Exception('cia for molecule %s already exists')

        if pair_filter is not None:
            if cia.pairName in pair_filter:
                self.log.info('Loading cia %s into model', cia.pairName)
                self.cia_dict[cia.pairName] = cia
        self.cia_dict[cia.pairName] = cia

    def load_cia_from_path(self, path, pair_filter=None):
        """
        Searches path for CIA files, creates and loads them into the cache
        ``.db`` will be loaded as :class:`~taurex.cia.picklecia.PickleCIA` and
        ``.cia`` files will be loaded as
        :class:`~taurex.cia.hitrancia.HitranCIA`

        Parameters
        ----------
        path : str
            Path to search for CIA files

        pair_filter : :obj:`list` of str , optional
            If provided, the cia will only be loaded
            if its pairname is in the list. Mostly used by the
            :func:`__getitem__` for filtering

        """
        from glob import glob
        from pathlib import Path
        import os
        from taurex.cia import PickleCIA

        # Find .db files
        glob_path = os.path.join(path, '*.db')

        file_list = glob(glob_path)
        self.log.debug('Glob list: %s', glob_path)
        self.log.debug('File list FOR CIA %s', file_list)

        for files in file_list:
            pairname = Path(files).stem.split('_')[0]

            self.log.debug('pairname found %s', pairname)

            if pair_filter is not None:
                if pairname not in pair_filter:
                    continue
            op = PickleCIA(files, pairname)
            self.add_cia(op)

        # Find .cia files
        glob_path = os.path.join(path, '*.cia')

        file_list = glob(glob_path)
        self.log.debug('File list %s', file_list)

        for files in file_list:
            from taurex.cia import HitranCIA
            pairname = Path(files).stem.split('_')[0]

            if pair_filter is not None:
                if pairname not in pair_filter:
                    continue
            op = HitranCIA(files)
            self.add_cia(op)

    def load_cia(self, cia_xsec=None, cia_path=None, pair_filter=None):
        """
        Main function to use when loading CIA files. Handles both
        cross sections and paths. Handles lists of either so lists of
        :class:`~taurex.cia.cia.CIA` objects or lists of paths can be used
        to load multiple files/objects


        Parameters
        ----------
        cia_xsec : :class:`~taurex.cia.cia.CIA` or :obj:`list` of :class:`~taurex.cia.cia.CIA` , optional
            Object(s) to include in cache

        cia_path : str or :obj:`list` of str, optional
            search path(s) to look for cias

        pair_filter : :obj:`list` of str , optional
            If provided, the cia will only be loaded
            if its pair name is in this list. Mostly used by the
            :func:`__getitem__` for filtering

        """

        from taurex.cia import CIA
        if cia_path is None:
            cia_path = self._cia_path

        self.log.debug('CIA XSEC, CIA_PATH %s %s', cia_xsec, cia_path)
        if cia_xsec is not None:
            if isinstance(cia_xsec, (list,)):
                self.log.debug('cia passed is list')
                for xsec in cia_xsec:
                    self.add_cia(xsec, pair_filter=pair_filter)
            elif isinstance(cia_xsec, CIA):
                self.add_cia(cia_xsec, pair_filter=pair_filter)
            else:
                self.log.error('Unknown type %s passed into cia, should be a list, single \
                     cia or None if reading a path', type(xsec))
                raise Exception('Unknown type passed into cia')
        if cia_path is not None:

            if isinstance(cia_path, str):
                self.load_cia_from_path(cia_path, pair_filter=pair_filter)
            elif isinstance(cia_path, (list,)):
                for path in cia_path:
                    self.load_cia_from_path(path, pair_filter=pair_filter)




"""
Contains the basic definition of an observed spectrum for TauRex 3
"""

from taurex.log import Logger
from taurex.output.writeable import Writeable


class BaseSpectrum(Logger, Writeable):
    """

    *Abstract class*

    A base class where spectrums are loaded (or later created). This
    is used to either plot against the forward model or passed into the 
    optimizer to be used to fit the forward model

    Parameters
    ----------

    name : str
        Name to be used in logging

    """

    def __init__(self, name):
        super().__init__(name)

    def create_binner(self):
        """
        Creates the appropriate binning object
        """
        from taurex.binning import FluxBinner

        return FluxBinner(wngrid=self.wavenumberGrid,
                          wngrid_width=self.binWidths)

    @property
    def spectrum(self):
        """
        **Requires Implementation**


        Should return the observed spectrum.

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def rawData(self):
        """
        **Requires Implementation**


        Should return the raw data set.

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def wavelengthGrid(self):
        """
        **Requires Implementation**


        Should return the wavelength grid of the spectrum in microns.
        This does not need to necessarily match the shape of :func:`spectrum`

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def wavenumberGrid(self):
        """
        Wavenumber grid in cm-1

        Returns
        -------
        wngrid : :obj:`array`

        """
        return 10000/self.wavelengthGrid

    @property
    def binEdges(self):
        """
        **Requires Implementation**


        Should return the bin edges of the wavenumber grid

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def binWidths(self):
        """
        **Requires Implementation**


        Should return the widths of each bin in the wavenumber grid

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def errorBar(self):
        """
        **Requires Implementation**


        Should return the error. *Must* be the same shape as
        :func:`spectrum`

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    def write(self, output):
        output.write_array('wlgrid', self.wavelengthGrid)
        output.write_array('spectrum', self.spectrum)
        output.write_array('binedges', self.binEdges)
        output.write_array('binwidths', self.binWidths)
        output.write_array('errorbars', self.errorBar)

        return output



from taurex.log import Logger

log = Logger('Factory')

def get_keywordarg_dict(klass):
    import inspect
    init_dicts = {}
    args, varargs, varkw, defaults = inspect.getargspec(klass.__init__)
    log.debug('Inpection {} {} {} {}'.format(args, varargs, varkw, defaults))
    if defaults is None:
        return init_dicts

    keyword_args = args[-len(defaults):]
    
    

    for keyword,value in zip(keyword_args,defaults):
        init_dicts[keyword] = value
    
    return init_dicts



def create_klass(config,klass):
    kwargs = get_keywordarg_dict(klass)

    for key in config:
        if key in kwargs:
            value = config[key]
            kwargs[key] = value
        else:
            log.error('Object {} does not have parameter {}'.format(klass.__name__,key))
            log.error('Available parameters are %s',kwargs.keys())
            raise KeyError
    obj = klass(**kwargs)

    # for key,value in config.items():
    #     try:
    #         obj[key] = value
    #     except KeyError:

    #         raise KeyError
    return obj





def create_profile(config, factory, baseclass=None):
    config, klass = determine_klass(config, 'profile_type',factory,
                                    baseclass)

    obj = create_klass(config,klass)
    
    return obj

def gas_factory(profile_type):
    if profile_type == 'constant':
        from taurex.data.profiles.chemistry import ConstantGas
        return ConstantGas
    elif profile_type in ('twopoint', '2point',):
        from taurex.data.profiles.chemistry import TwoPointGas
        return TwoPointGas
    elif profile_type in ('twolayer','2layer',):
        from taurex.data.profiles.chemistry import TwoLayerGas
        return TwoLayerGas
    else:
        raise NotImplementedError('Gas profile {} not implemented'.format(profile_type))



def temp_factory(profile_type):
    if profile_type == 'isothermal':
        from taurex.data.profiles.temperature import Isothermal
        return Isothermal
    elif profile_type in ('guillot','guillot2010',):
        from taurex.data.profiles.temperature import Guillot2010
        return Guillot2010
    elif profile_type in ('npoint',):
        from taurex.data.profiles.temperature import NPoint
        return NPoint
    elif profile_type in ('rodgers','rodgers2010',):
        from taurex.data.profiles.temperature import Rodgers2000
        return Rodgers2000
    elif profile_type in ('file',):
        from taurex.data.profiles.temperature import TemperatureFile
        return TemperatureFile
    else:
        raise NotImplementedError('Temperature profile {} not implemented'.format(profile_type))

def pressure_factory(profile_type):
    if profile_type == 'simple':
        from taurex.data.profiles.pressure import SimplePressureProfile
        return SimplePressureProfile
    else:
        raise NotImplementedError('Pressure profile {} not implemented'.format(profile_type))

def star_factory(star_type):
    if star_type == 'blackbody':
        from taurex.data.stellar import BlackbodyStar
        return BlackbodyStar
    elif star_type == 'phoenix':
        from taurex.data.stellar import PhoenixStar
        return PhoenixStar
    else:
        raise NotImplementedError('Star of type {} not implemented'.format(star_type))


def create_gas_profile(config):
    from taurex.data.profiles.chemistry.gas import Gas
    return create_profile(config, gas_factory, Gas)


def create_temperature_profile(config):
    from taurex.data.profiles.temperature import TemperatureProfile
    return create_profile(config, temp_factory, TemperatureProfile)


def create_pressure_profile(config):
    from taurex.data.profiles.pressure.pressureprofile import PressureProfile
    return create_profile(config, pressure_factory, PressureProfile)


def create_ace(config):
    from taurex.data.profiles.chemistry import ACEChemistry
    return ACEChemistry(**config)


def create_chemistry(config):
    try:
        chemistry = config.pop('chemistry_type').lower()
    except KeyError:
        log.error('No chemistry defined in input')
        raise KeyError    

    if chemistry in ('ace','equilibrium'):
        return create_ace(config)
    elif chemistry in ('file', ):
        from taurex.chemistry import ChemistryFile
        return create_klass(config, ChemistryFile)
    elif chemistry in ('custom',):
        from taurex.chemistry import Chemistry
        config['chemistry_type'] = 'custom'
        config, klass = determine_klass(config, 'chemistry_type', None,
                        Chemistry)
        obj = klass(**config)
        return obj
    elif chemistry in ('taurex', 'complex', 'defined', 'free'):
        from taurex.data.profiles.chemistry import TaurexChemistry
        gases = []
        config_key=[]
        for key, value in config.items():
            
            if isinstance(value, dict):
                log.debug('FOUND GAS {} {}'.format(key, value))
                config_key.append(key)
                gas_type = value.pop('gas_type').lower()
                klass = gas_factory(gas_type)
                gases.append(klass(molecule_name=key, **value))

        for k in config_key: del config[k]
        log.debug('leftover keys {}'.format(config))
        obj = TaurexChemistry(**config)
        for g in gases:
            obj.addGas(g)
    
        return obj
    else:
        raise ValueError('Unknown chemistry type {}'.format(chemistry))


def model_factory(model_type):
    if model_type == 'transmission':
        from taurex.model import TransmissionModel
        return TransmissionModel
    elif model_type == 'emission':
        from taurex.model import EmissionModel
        return EmissionModel
    elif model_type in ('directimage', 'direct image'):
        from taurex.model import DirectImageModel
        return DirectImageModel
    else:
        raise NotImplementedError('Model {} not implemented'.format(model_type))


def planet_factory(planet_type):
    if planet_type in ('simple', 'planet', 'basic', 'meme',):
        from taurex.data import Planet
        return Planet
    else:
        raise NotImplementedError('Model {} not implemented'.format(model_type))

def optimizer_factory(optimizer):
    if optimizer == 'nestle':
        from taurex.optimizer.nestle import NestleOptimizer
        return NestleOptimizer
    elif optimizer in ('multinest','pymultinest',):
        from taurex.optimizer.multinest import MultiNestOptimizer
        return MultiNestOptimizer
    elif optimizer in ('polychord','pypolychord'):
        from taurex.optimizer.polychord import PolyChordOptimizer
        return PolyChordOptimizer
    elif optimizer in ('dypolychord','dynamic-polychord'):
        from taurex.optimizer.dypolychord import dyPolyChordOptimizer
        return dyPolyChordOptimizer
    else:
        raise NotImplementedError('Optimizer {} not implemented'.format(optimizer))    


def instrument_factory(instrument):
    if instrument in ('file', 'fromfile',):
        from taurex.instruments import InstrumentFile
        return InstrumentFile


def create_star(config):
    from taurex.data.stellar.star import BlackbodyStar
    config, klass = determine_klass(config, 'star_type', star_factory,
                                    BlackbodyStar)

    obj = klass(**config)

    return obj


def create_planet(config):
    from taurex.data.planet import Planet

    if 'planet_type' not in config:
        config['planet_type'] = 'simple'

    config, klass = determine_klass(config, 'planet_type', planet_factory,
                                    Planet)

    obj = klass(**config)

    return obj


def create_optimizer(config):
    from taurex.optimizer.optimizer import Optimizer
    config, klass = determine_klass(config, 'optimizer', optimizer_factory,
                                    Optimizer)

    obj = klass(**config)
    
    return obj


def determine_klass(config, field, factory, baseclass=None):

    try:
        klass_field = config.pop(field).lower()
    except KeyError:
        log.error('Field not defined in {}'.format(field))
        raise KeyError

    klass = None
    if klass_field == 'custom':
        try:
            python_file = config.pop('python_file').lower()
        except KeyError:
            log.error('No python file for custom profile/model')
            raise KeyError

        klass = detect_and_return_klass(python_file, baseclass)
    else:
        klass = factory(klass_field)

    return config, klass


def create_instrument(config):
    from taurex.instruments.instrument import Instrument
    config, klass = determine_klass(config, 'instrument', instrument_factory,
                                    Instrument)

    obj = klass(**config)
    
    return obj

def generate_contributions(config):
    from taurex.contributions import AbsorptionContribution, CIAContribution, RayleighContribution

    contributions = []
    for key in config.keys():
        if key == 'Absorption':
            contributions.append(create_klass(config[key],AbsorptionContribution))
        elif key == 'CIA':
            contributions.append(create_klass(config[key],CIAContribution))
        elif key == 'Rayleigh':
            contributions.append(create_klass(config[key],RayleighContribution))
        elif key == 'SimpleClouds':
             from taurex.contributions import SimpleCloudsContribution
             contributions.append(create_klass(config[key],SimpleCloudsContribution))
        elif key == 'BHMie':
             from taurex.contributions import BHMieContribution
             contributions.append(create_klass(config[key],BHMieContribution))
        elif key == 'LeeMie':
             from taurex.contributions import LeeMieContribution
             contributions.append(create_klass(config[key],LeeMieContribution))
        elif key == 'FlatMie':
             from taurex.contributions import FlatMieContribution
             contributions.append(create_klass(config[key],FlatMieContribution))

    return contributions

    
def create_model(config,gas,temperature,pressure,planet,star):
    from taurex.model import ForwardModel
    log.debug(config)
    config, klass = determine_klass(config, 'model_type', model_factory,
                                    ForwardModel)

    log.debug('Chosen_model is {}'.format(klass))
    kwargs = get_keywordarg_dict(klass)
    log.debug('Model kwargs {}'.format(kwargs))
    log.debug('---------------{} {}--------------'.format(gas,gas.activeGases))
    if 'planet' in kwargs:
        kwargs['planet'] = planet
    if 'star' in kwargs:
        kwargs['star'] = star
    if 'chemistry' in kwargs:
        kwargs['chemistry'] = gas
    if 'temperature_profile' in kwargs:
        kwargs['temperature_profile'] =temperature
    if 'pressure_profile' in kwargs:
        kwargs['pressure_profile'] = pressure
    log.debug('New Model kwargs {}'.format(kwargs))
    log.debug('Creating model---------------')
    
    kwargs.update(dict([(k,v) for k,v in config.items() if not isinstance(v,dict)]))
    obj = klass(**kwargs)
    
    contribs = generate_contributions(config)

    for c in contribs:
        obj.add_contribution(c)


    return obj       

def detect_and_return_klass(python_file, baseclass):
    import importlib.util 
    import inspect
    spec = importlib.util.spec_from_file_location("foo", python_file) 
    foo = importlib.util.module_from_spec(spec) 
    spec.loader.exec_module(foo) 
    classes = [m[1] for m in inspect.getmembers(foo, inspect.isclass) if m[1] \
               is not baseclass and issubclass(m[1],baseclass)]

    if len(classes) == 0:
        self.error('Could not find class of type %s in file %s',baseclass, python_file)
        raise Exception(f'No class inheriting from {baseclass} in '
                        f'{python_file}')
    return classes[0]



from .tprofile import TemperatureProfile
import numpy as np
from taurex.data.fittable import fitparam


class Isothermal(TemperatureProfile):
    """An isothermal temperature-pressure profile

    Parameters
    ----------

    T : float
        Isothermal Temperature to set

    """

    def __init__(self, T=1500):
        super().__init__('Isothermal')

        self._iso_temp = T

    @fitparam(param_name='T',
              param_latex='$T$',
              default_fit=False,
              default_bounds=[300.0, 2000.0])
    def isoTemperature(self):
        """Isothermal temperature in Kelvin"""
        return self._iso_temp

    @isoTemperature.setter
    def isoTemperature(self, value):
        self._iso_temp = value

    @property
    def profile(self):
        """Returns an isothermal temperature profile

        Returns: :obj:`array`
            temperature profile
        """

        T = np.zeros((self.nlayers))
        T[:] = self._iso_temp

        return T

    def write(self, output):
        temperature = super().write(output)
        temperature.write_scalar('T', self._iso_temp)
        return temperature




from .contribution import Contribution
import numpy as np
from taurex.cache import OpacityCache


class AbsorptionContribution(Contribution):
    """
    Computes the contribution to the optical depth
    occuring from molecular absorption.
    """

    def __init__(self):
        super().__init__('Absorption')
        self._opacity_cache = OpacityCache()

    def prepare_each(self, model, wngrid):
        """
        Prepares each molecular opacity by weighting them
        by their mixing ratio in the atmosphere

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            Name of molecule and weighted opacity

        """

        self.debug('Preparing model with %s', wngrid.shape)
        self._ngrid = wngrid.shape[0]
        sigma_xsec = np.zeros(shape=(model.nLayers, wngrid.shape[0]))

        # Get the opacity cache
        self._opacity_cache = OpacityCache()
        # Loop through all active gases
        for gas in model.chemistry.activeGases:

            # Clear sigma array
            sigma_xsec[...] = 0.0

            # Get the mix ratio of the gas
            gas_mix = model.chemistry.get_gas_mix_profile(gas)
            self.info('Recomputing active gas %s opacity', gas)

            # Get the cross section object relating to the gas
            xsec = self._opacity_cache[gas]
            # Loop through the layers
            for idx_layer, tp in enumerate(zip(model.temperatureProfile,
                                               model.pressureProfile)):
                self.debug('Got index,tp %s %s', idx_layer, tp)

                temperature, pressure = tp

                # Place into the array
                sigma_xsec[idx_layer] += \
                    xsec.opacity(temperature, pressure,
                                 wngrid)*gas_mix[idx_layer]

            # Temporarily assign to master cross-section
            self.sigma_xsec = sigma_xsec
            yield gas, sigma_xsec

    @property
    def sigma(self):
        """
        Returns the fused weighted cross-section
        of all active gases
        """
        return self.sigma_xsec



from .chemistry import Chemistry
from taurex.external.ace import md_ace
from taurex.data.fittable import fitparam
import numpy as np
import math
from taurex.cache import OpacityCache


class ACEChemistry(Chemistry):
    """
    Equilibrium chemistry
    Computes chemical profile using the Aerotherm Chemical Equilibrium (ACE)
    Fortran code by
    Agúndez, M., Venot, O., Iro, N., et al. 2012, AandA, 548,A73

    Parameters
    ----------
    metallicity : float
        Stellar metallicity in solar units

    co_ratio : float
        C/O ratio

    therm_file : str , optional
        Location of NASA.therm file. If not set will use file included
        in library

    spec_file : str , optional
        Location of composes.dat.  If not set will use file included in library


    """

    ace_H_solar = 12.0
    """H solar abundance"""
    ace_He_solar = 10.93
    """He solar abundance"""
    ace_C_solar = 8.43
    """C solar abundance"""
    ace_O_solar = 8.69
    """O solar abundance"""
    ace_N_solar = 7.83
    """N solar abundance"""

    def __init__(self, ace_metallicity=1.0,
                 ace_co_ratio=0.54951,
                 therm_file=None,
                 spec_file=None):

        super().__init__('ACE')
        self.ace_metallicity = ace_metallicity
        self.ace_co = ace_co_ratio
        self.active_gases = None
        self.inactive_gases = None
        self._get_files(therm_file, spec_file)
        self.active_mixratio_profile = None
        self.inactive_mixratio_profile = None

    @property
    def activeGases(self):
        """
        Returns names of actively absorbing molecules

        Returns
        -------

        active : :obj:`list` of str
            List of molecules

        """
        return self.active_gases

    @property
    def inactiveGases(self):
        """
        Returns names of molecules not actively absorbing

        Returns
        -------

        inactive : :obj:`list` of str
            List of molecules

        """
        return self.inactive_gases

    @property
    def activeGasMixProfile(self):
        """
        Layer by layer mixing ratio of ``active`` molecules

        Returns
        -------

        mix_profile :  :obj:`array`
            Array of shape ``(nactivegases,nlayers)``


        """
        return self.active_mixratio_profile

    @property
    def inactiveGasMixProfile(self):
        """
        Layer by layer mixing ratio of ``inactive`` molecules

        Returns
        -------

        mix_profile :  :obj:`array`
            Array of shape ``(ninactivegases,nlayers)``


        """
        return self.inactive_mixratio_profile

    def _get_files(self, therm_file, spec_file):
        import os
        import taurex.external

        path_to_files = os.path.join(os.path.abspath(
            os.path.dirname(taurex.external.__file__)), 'ACE')
        self._specfile = spec_file
        self._thermfile = therm_file
        if self._specfile is None:
            self._specfile = os.path.join(path_to_files, 'composes.dat')
        if self._thermfile is None:
            self._thermfile = os.path.join(path_to_files, 'NASA.therm')

    def _get_gas_mask(self):
        import operator
        from taurex.constants import AMU
        self._active_mask = np.ndarray(shape=(105,), dtype=np.bool)
        self._inactive_mask = np.ndarray(shape=(105,), dtype=np.bool)
        self._active_mask[:] = False
        self._inactive_mask[:] = False
        new_active_gases = []
        new_inactive_gases = []

        self._molecule_weight = {}

        with open(self._specfile, 'r') as textfile:

            molecules_i_have = OpacityCache().find_list_of_molecules()
            self.debug('MOLECULES %s', molecules_i_have)
            for line in textfile:
                sl = line.split()
                idx = int(sl[0])-1
                molecule = sl[1]
                self._molecule_weight[molecule] = float(sl[2])*AMU

                if molecule in molecules_i_have:
                    self._active_mask[idx] = True
                    new_active_gases.append((molecule, idx))
                else:
                    self._inactive_mask[idx] = True
                    new_inactive_gases.append((molecule, idx))
        # Create a new list where the gases are in the correct order
        new_active_gases.sort(key=operator.itemgetter(1))
        new_inactive_gases.sort(key=operator.itemgetter(1))

        self.active_gases = [molecule for molecule, _ in new_active_gases]
        self.inactive_gases = [molecule for molecule, _ in new_inactive_gases]
        self.info('Active gases: %s', self.active_gases)
        self.info('Inactive gases: %s', self.inactive_gases)

    def set_ace_params(self):

        # set O, C and N abundances given metallicity (in solar units) and CO
        self.O_abund_dex = math.log10(self.ace_metallicity *
                                      (10**(self.ace_O_solar-12.)))+12.
        self.N_abund_dex = math.log10(self.ace_metallicity *
                                      (10**(self.ace_N_solar-12.)))+12.

        self.C_abund_dex = self.O_abund_dex + math.log10(self.ace_co)

        # H and He don't change
        self.H_abund_dex = self.ace_H_solar
        self.He_abund_dex = self.ace_He_solar

    def compute_active_gas_profile(self, nlayers, altitude_profile,
                                   pressure_profile,
                                   temperature_profile):
        """Computes gas profiles of both active and inactive molecules for each layer

        Parameters
        ----------

        altitude_profile : array_like
            Altitude profile of atmosphere (usually computed in model)

        pressure_profile : array_like
            Pressure profile of atmosphere

        temperature_profile : array_like
            Temperature profile of atmosphere

        """

        self._get_gas_mask()
        self.active_mixratio_profile = np.zeros(shape=(len(self.activeGases),
                                                       nlayers))
        self.inactive_mixratio_profile = np.zeros((len(self.inactiveGases),
                                                   nlayers))
        self.set_ace_params()

        # Call FORTRAN ACE function
        self._ace_profile = md_ace(self._specfile,
                                   self._thermfile,
                                   altitude_profile/1000.0,
                                   pressure_profile/1.e5,
                                   temperature_profile,
                                   self.He_abund_dex,
                                   self.C_abund_dex,
                                   self.O_abund_dex,
                                   self.N_abund_dex)

        self.active_mixratio_profile = self._ace_profile[self._active_mask, :]
        self.inactive_mixratio_profile = \
            self._ace_profile[self._inactive_mask, :]

    def initialize_chemistry(self, nlayers=100, temperature_profile=None,
                             pressure_profile=None, altitude_profile=None):
        """
        Sets up and and constructs chemical profiles. Called by forward
        model before path calculation

        Parameters
        ----------

        nlayers : int
            Number of layers in atmosphere

        altitude_profile : array_like
            Altitude profile of atmosphere (usually computed in model)

        pressure_profile : array_like
            Pressure profile of atmosphere

        temperature_profile : array_like
            Temperature profile of atmosphere

        """

        self.info('Initializing chemistry model')

        self.compute_active_gas_profile(nlayers, altitude_profile,
                                        pressure_profile, temperature_profile)

        self.compute_mu_profile(nlayers)

    @fitparam(param_name='ace_metallicity',
              param_latex='Metallicity',
              default_mode='log',
              default_fit=False,
              default_bounds=[-1, 4])
    def aceMetallicity(self):
        """
        Metallicity of star in solar units
        """
        return self.ace_metallicity

    @aceMetallicity.setter
    def aceMetallicity(self, value):
        self.ace_metallicity = value

    @fitparam(param_name='ace_co',
              param_latex='C/O',
              default_fit=False,
              default_bounds=[0, 2])
    def aceCORatio(self):
        """
        CO ratio of star
        """
        return self.ace_co

    @aceCORatio.setter
    def aceCORatio(self, value):
        self.ace_co = value

    def write(self, output):

        gas_entry = super().write(output)
        gas_entry.write_scalar('ace_metallicity', self.ace_metallicity)
        gas_entry.write_scalar('ace_co_ratio', self.ace_co)
        if self._thermfile is not None:
            gas_entry.write_string('therm_file', self._thermfile)
        if self._specfile is not None:
            gas_entry.write_string('spec_file', self._specfile)

        return gas_entry

    def compute_mu_profile(self, nlayers):
        self.mu_profile = np.zeros(shape=(nlayers,))
        if self.activeGasMixProfile is not None:
            for idx, gasname in enumerate(self.activeGases):
                self.mu_profile += self.activeGasMixProfile[idx, :] * \
                    self._molecule_weight[gasname]
        if self.inactiveGasMixProfile is not None:
            for idx, gasname in enumerate(self.inactiveGases):
                self.mu_profile += self.inactiveGasMixProfile[idx, :] * \
                    self._molecule_weight[gasname]



import logging
from taurex.log import Logger
import numpy as np
from taurex.output.writeable import Writeable


class LightCurveData(Logger, Writeable):

    availableInstruments = ['wfc3', 'spitzer', 'stis', 'twinkle']

    @classmethod
    def fromInstrumentName(cls, name, lc_data):
        log = logging.getLogger()
        if name.lower() in ('wfc3',):
            return cls(lc_data, 'wfc3', (1.1, 1.8))
        elif name.lower() in ('spitzer',):
            return cls(lc_data, 'spitzer', (3.4, 8.2))
        elif name.lower() in ('stis',):
            return cls(lc_data, 'stis', (0.3, 1.0))
        elif name.lower() in ('twinkle',):
            return cls(lc_data, 'twinkle', (0.4, 4.5))
        else:
            log.error(
                'LightCurve of instrument %s not recognized'
                ' or implemented', name)
            raise KeyError

    def __init__(self, lc_data, instrument_name, wavelength_region):
        super().__init__(self.__class__.__name__)
        self._instrument_name = instrument_name
        # new version
        if self._instrument_name not in lc_data:
            self.error('Instrument with key %s not found in pickled lightcurve'
                       ' file', self._instrument_name)
            raise KeyError()

        self._wavelength_region = wavelength_region
        self._load_data(lc_data)

    def _load_data(self, lc_data):
        # new
        self._time_series = lc_data[self._instrument_name]['time_series']

        # new
        self._raw_data = lc_data[self._instrument_name]['data'][:, :, 0]
        self._data_std = lc_data[self._instrument_name]['data'][:, :, 1]

        self._max_nfactor = np.max(self._raw_data, axis=1)
        self._min_nfactor = np.min(self._raw_data, axis=1)

    @property
    def instrumentName(self):
        return self._instrument_name

    @property
    def wavelengthRegion(self):
        return self._wavelength_region

    @property
    def timeSeries(self):
        return self._time_series

    @property
    def rawData(self):
        return self._raw_data

    @property
    def dataError(self):
        return self._data_std

    @property
    def minNFactors(self):
        return self._min_nfactor

    @property
    def maxNFactors(self):
        return self._max_nfactor

    def write(self, output):

        lc_grp = output.create_group(self.instrumentName)

        lc_grp.write_array('raw_data', self.rawData)
        lc_grp.write_array('data_error', self.dataError)
        lc_grp.write_array('min_n_factors', self.minNFactors)
        lc_grp.write_array('max_n_factors', self.maxNFactors)
        lc_grp.write_array('time_series', self.timeSeries)
        lc_grp.write_scalar('min_wavelength', self.wavelengthRegion[0])
        lc_grp.write_scalar('max_wavelength', self.wavelengthRegion[1])

        return lc_grp



from .gas import Gas
from taurex.util import molecule_texlabel
import math
import numpy as np


class TwoPointGas(Gas):
    """

    Two point gas profile.

    A gas profile with two different mixing layers at the surface of the
    planet and top of the atmosphere and interpolated between the two

    Parameters
    -----------
    molecule_name : str
        Name of molecule

    mix_ratio_surface : float
        Mixing ratio of the molecule on the planet surface

    mix_ratio_top : float
        Mixing ratio of the molecule at the top of the atmosphere


    """

    def __init__(self, molecule_name='CH4',
                 mix_ratio_surface=1e-4, mix_ratio_top=1e-8):
        super().__init__('TwoPointGas', molecule_name)
        self._mix_surface = mix_ratio_surface
        self._mix_top = mix_ratio_top
        self.add_surface_param()
        self.add_top_param()
        self._mix_profile = None

    @property
    def mixProfile(self):
        """

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer
        """
        return self._mix_profile

    @property
    def mixRatioSurface(self):
        """Abundance on the planets surface"""
        return self._mix_surface

    @property
    def mixRatioTop(self):
        """Abundance on the top of atmosphere"""
        return self._mix_top

    @mixRatioSurface.setter
    def mixRatioSurface(self, value):
        self._mix_surface = value

    @mixRatioTop.setter
    def mixRatioTop(self, value):
        self._mix_top = value

    def add_surface_param(self):

        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_surface = '{}_surface'.format(param_name)
        param_surf_tex = '{}_surface'.format(param_tex)

        def read_surf(self):
            return self._mix_surface

        def write_surf(self, value):
            self._mix_surface = value

        fget_surf = read_surf
        fset_surf = write_surf

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(param_surface, param_surf_tex, fget_surf,
                                fset_surf, 'log', default_fit, bounds)

    def add_top_param(self):

        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_top = '{}_top'.format(param_name)
        param_top_tex = '{}_top'.format(param_tex)

        def read_top(self):
            return self._mix_top

        def write_top(self, value):
            self._mix_top = value

        fget_top = read_top
        fset_top = write_top

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(param_top, param_top_tex, fget_top, fset_top,
                                'log', default_fit, bounds)

    def initialize_profile(self, nlayers=None, temperature_profile=None,
                           pressure_profile=None, altitude_profile=None):
        self._mix_profile = np.zeros(nlayers)

        chem_surf = self._mix_surface
        chem_top = self._mix_top
        p_surf = pressure_profile[0]
        p_top = pressure_profile[-1]

        a = (math.log10(chem_surf)-math.log10(chem_top)) / \
            (math.log10(p_surf)-math.log10(p_top))

        b = math.log10(chem_surf)-a*math.log10(p_surf)

        self._mix_profile[1:-1] = 10**(a * np.log10(pressure_profile[1:-1])+b)
        self._mix_profile[0] = chem_surf
        self._mix_profile[-1] = chem_top

    def write(self, output):
        gas_entry = super().write(output)
        gas_entry.write_scalar('mix_ratio_top', self.mixRatioTop)
        gas_entry.write_scalar('mix_ratio_surface', self.mixRatioSurface)
        return gas_entry



from .contribution import Contribution
import numpy as np
import numba
from taurex.cache import CIACache


@numba.jit(nopython=True, nogil=True)
def contribute_cia(startK, endK, density_offset, sigma, density, path, nlayers,
                   ngrid, layer, tau):
    """
    Collisionally induced absorption integration function

    This has the form:

    .. math::

        \\tau_{\\lambda}(z) = \\int_{z_{0}}^{z_{1}}
            \\sigma(z') \\rho(z')^{2} dz',

    where :math:`z` is the layer, :math:`z_0` and :math:`z_1` are ``startK``
    and ``endK`` respectively. :math:`\\sigma` is the weighted
    cross-section ``sigma``. :math:`rho` is the ``density`` and
    :math:`dz'` is the integration path length ``path``


    Parameters
    ----------
    startK: int
        starting layer in integration

    endK: int
        last layer in integration

    density_offset: int
        Which part of the density profile to start from

    sigma: :obj:`array`
        cross-section

    density: array_like
        density profile of atmosphere

    path: array_like
        path-length or altitude gradient

    nlayers: int
        Total number of layers (unused)

    ngrid: int
        total number of grid points

    layer: int
        Which layer we currently on

    Returns
    -------
    tau : array_like
        optical depth (well almost you still need to do ``exp(-tau)`` yourself)

    """
    for k in range(startK, endK):
        _path = path[k]
        _density = density[k+density_offset]
        # for mol in range(nmols):
        for wn in range(ngrid):
            tau[layer, wn] += sigma[k+layer, wn]*_path*_density*_density


class CIAContribution(Contribution):
    """
    Computes the contribution to the optical depth
    occuring from collisionally induced absorption.

    Parameters
    ----------
    cia_pairs: :obj:`list` of str
        list of molecule pairs of the form ``mol1-mol2``
        e.g. ``H2-He``
    """
    def __init__(self, cia_pairs=None):
        super().__init__('CIA')
        self._cia_pairs = cia_pairs

        self._cia_cache = CIACache()
        if self._cia_pairs is None:
            self._cia_pairs = []

    @property
    def ciaPairs(self):
        """
        Returns list of molecular pairs involved

        Returns
        -------
        :obj:`list` of str
        """

        return self._cia_pairs

    @ciaPairs.setter
    def ciaPairs(self, value):
        self._cia_pairs = value

    def contribute(self, model, start_layer, end_layer, density_offset, layer,
                   density, tau, path_length=None):
        if self._total_cia > 0:
            contribute_cia(start_layer, end_layer, density_offset,
                           self.sigma_xsec, density, path_length,
                           self._nlayers, self._ngrid,
                           layer, tau)

    def prepare_each(self, model, wngrid):
        """
        Computes and weighs cross-section for
        a single pair of molecules

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid


        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            Molecular pair and the weighted cia opacity.

        """

        self._total_cia = len(self.ciaPairs)
        self._nlayers = model.nLayers
        self._ngrid = wngrid.shape[0]
        self.info('Computing CIA ')

        sigma_cia = np.zeros(shape=(model.nLayers, wngrid.shape[0]))

        chemistry = model.chemistry

        for pairName in self.ciaPairs:
            cia = self._cia_cache[pairName]
            sigma_cia[...] = 0.0

            cia_factor = chemistry.get_gas_mix_profile(cia.pairOne) * \
                chemistry.get_gas_mix_profile(cia.pairTwo)

            for idx_layer, temperature in enumerate(model.temperatureProfile):
                _cia_xsec = cia.cia(temperature, wngrid)
                sigma_cia[idx_layer] += _cia_xsec*cia_factor[idx_layer]
            self.sigma_xsec = sigma_cia
            yield pairName, sigma_cia

    def write(self, output):
        contrib = super().write(output)
        if len(self.ciaPairs) > 0:
            contrib.write_string_array('cia_pairs', self.ciaPairs)
        return contrib



"""These modules deal with binning down results from models"""
from .binner import Binner
from .simplebinner import SimpleBinner
from .fluxbinner import FluxBinner
from .nativebinner import NativeBinner



from .array import ArraySpectrum
import numpy as np
import pickle


class IraclisSpectrum(ArraySpectrum):
    """
    Loads an observation from Iraclis pickle data

    Parameters
    -----------
    filename: string
        Path to observed spectrum file.

    """

    def __init__(self, filename):
        self._filename = filename
        try:
            with open(filename, 'rb') as f:
                database = pickle.load(f)
        except UnicodeDecodeError:
            with open(filename, 'rb') as f:
                database = pickle.load(f, encoding='latin1')

        wl = database['spectrum']['wavelength']
        td = database['spectrum']['depth']
        err = database['spectrum']['error']
        width = database['spectrum']['width']

        final_array = np.vstack((wl, td, err, width)).T

        super().__init__(final_array)





from .data.profiles.chemistry import ChemistryFile
from .data.profiles.chemistry.chemistry import Chemistry
from .data.profiles.chemistry import TaurexChemistry
try:
    from .data.profiles.chemistry.acechemistry import ACEChemistry
except ImportError:
    print('Ace library not found. ACEChemistry could not be loaded')
from .data.profiles.chemistry.gas.gas import Gas
from .data.profiles.chemistry import ConstantGas
from .data.profiles.chemistry import TwoLayerGas



"""
Module dealing with observed lightcurves
"""

from .spectrum import BaseSpectrum
import numpy as np
from taurex.model.lightcurve.lightcurvedata import LightCurveData


class ObservedLightCurve(BaseSpectrum):
    """

    Loads an observed lightcurve from a pickle file.

    Parameters
    ----------

    filename : str
        Path to pickle file containing lightcurve data

    """

    def __init__(self, filename):
        super().__init__('observed_lightcurve')

        import pickle
        with open(filename, 'rb') as f:
            lc_data = pickle.load(f, encoding='latin1')
        # new version
        self.obs_spectrum = \
            np.empty(shape=(len(lc_data['obs_spectrum'][:, 0]), 4))

        # new version
        self.obs_spectrum[:, 0] = lc_data['obs_spectrum'][:, 0]
        self.obs_spectrum[:, 1] = lc_data['obs_spectrum'][:, 1]
        self.obs_spectrum[:, 2] = lc_data['obs_spectrum'][:, 2]
        self.obs_spectrum[:, 3] = lc_data['obs_spectrum'][:, 3]

        self._spec, self._std = self._load_data_file(lc_data)

    def create_binner(self):
        from taurex.binning.lightcurvebinner import LightcurveBinner
        """
        Creates the appropriate binning object
        """

        return LightcurveBinner()

    def _load_data_file(self, lc_data):
        """load data from different instruments."""

        raw_data = []
        data_std = []
        wngrid_min = []

        for i in LightCurveData.availableInstruments:
            # new version
            # raw data includes data and datastd.
            if i in lc_data:
                wngrid_min.append(lc_data[i]['wl_grid'].min())
                raw_data.append(lc_data[i]['data'][:, :, 0])
                data_std.append(lc_data[i]['data'][:, :, 1])
        wngrid_min, raw_data, data_std = \
            (list(t) for t in zip(*sorted(zip(wngrid_min, raw_data, data_std),
                                          key=lambda x: x[0], reverse=True)))
        return np.concatenate(raw_data), np.concatenate(data_std)

    @property
    def spectrum(self):
        """
        Returns Light curve spectrum.
        The lightcurve spectrum comes in the form of multiple lightcurves
        stuck together into
        one long spectrum. The number of lightcurves is equal to the number of
        bins in :func:`wavelengthGrid`.

        Returns
        -------
        spectrum : :obj:`array`

        """
        return self._spec

    @property
    def rawData(self):
        """
        Raw lightcurve data read from file

        Returns
        -------
        lc_data : :obj:`array`

        """
        self.obs_spectrum

    @property
    def wavelengthGrid(self):
        """
        Returns wavelength grid in microns

        Returns
        -------
        wlgrid : :obj:`array`

        """
        return self.obs_spectrum[:, 0]

    @property
    def binEdges(self):
        """
        Returns bin edges for wavelength grid

        Returns
        -------
        out : :obj:`array`

        """
        return self.obs_spectrum[:, 3]

    @property
    def binWidths(self):
        """
        Widths for each bin in wavelength grid

        Returns
        -------
        out : :obj:`array`

        """
        return np.zeros(2)

    @property
    def errorBar(self):
        """
        Like :func:`spectrum` except its the error at each point in the
        lightcurve spectrum

        Returns
        -------
        err : :obj:`array`
            Error at each point in lightcurve spectrum

        """
        return self._std

    def write(self, output):
        output.write_array('wlgrid', self.wavelengthGrid)
        output.write_array('spectrum', self.obs_spectrum[:, 1])
        output.write_array('lightcurve', self.spectrum)
        output.write_array('binedges', self.binEdges)
        output.write_array('binwidths', self.binWidths)
        output.write_array('errorbars', self.obs_spectrum[:, 2])
        output.write_array('lightcurve_errorbars', self.errorBar)

        return output



from .instrument import Instrument
import numpy as np
from taurex.binning import FluxBinner
import math
from taurex.util.util import wnwidth_to_wlwidth


class InstrumentFile(Instrument):
    """
    Loads a 2-3 column file with wlgrid
    and noise and maybe wlgrid


    Parameters
    ----------

    filename: str
        Filename of file containing the binning and error


    """

    def __init__(self, filename=None):
        super().__init__()

        self._spectrum = np.loadtxt(filename)

        self._wlgrid = self._spectrum[:, 0]
        
        sortedwl = self._wlgrid.argsort()[::-1]

        self._wlgrid = self._wlgrid[sortedwl]

        self._wngrid = 10000/self._wlgrid

        self._noise = self._spectrum[sortedwl, 1]

        try:
            self._wlwidths = self._spectrum[sortedwl, 2]
        except IndexError:
            from taurex.util.util import compute_bin_edges

            self._wlwidths - compute_bin_edges(self._wlgrid)[-1]

        self.create_wn_widths()

        self._binner = FluxBinner(self._wngrid, wngrid_width=self._wnwidths)

    def create_wn_widths(self):

        self._wnwidths = wnwidth_to_wlwidth(self._wlgrid, self._wlwidths)

    def model_noise(self, model, model_res=None, num_observations=1):

        if model_res is None:
            model_res = model.model()

        wngrid, spectrum, error, grid_width = self._binner.bin_model(model_res)

        return wngrid, spectrum, self._noise / math.sqrt(num_observations), grid_width



import numpy as np
from .simplemodel import SimpleForwardModel


class TransmissionModel(SimpleForwardModel):
    """

    A forward model for transits

    Parameters
    ----------

    planet: :class:`~taurex.data.planet.Planet`, optional
        Planet model, default planet is Jupiter

    star: :class:`~taurex.data.stellar.star.Star`, optional
        Star model, default star is Sun-like

    pressure_profile: :class:`~taurex.data.profiles.pressure.pressureprofile.PressureProfile`, optional
        Pressure model, alternative is to set ``nlayers``, ``atm_min_pressure``
        and ``atm_max_pressure``

    temperature_profile: :class:`~taurex.data.profiles.temperature.tprofile.TemperatureProfile`, optional
        Temperature model, default is an :class:`~taurex.data.profiles.temperature.isothermal.Isothermal`
        profile at 1500 K

    chemistry: :class:`~taurex.data.profiles.chemistry.chemistry.Chemistry`, optional
        Chemistry model, default is
        :class:`~taurex.data.profiles.chemistry.taurexchemistry.TaurexChemistry` with
        ``H2O`` and ``CH4``

    nlayers: int, optional
        Number of layers. Used if ``pressure_profile`` is not defined.

    atm_min_pressure: float, optional
        Pressure at TOA. Used if ``pressure_profile`` is not defined.

    atm_max_pressure: float, optional
        Pressure at BOA. Used if ``pressure_profile`` is not defined.

    """

    def __init__(self,
                 planet=None,
                 star=None,
                 pressure_profile=None,
                 temperature_profile=None,
                 chemistry=None,
                 nlayers=100,
                 atm_min_pressure=1e-4,
                 atm_max_pressure=1e6):

        super().__init__(self.__class__.__name__, planet,
                         star,
                         pressure_profile,
                         temperature_profile,
                         chemistry,
                         nlayers,
                         atm_min_pressure,
                         atm_max_pressure)

    def compute_path_length(self, dz):

        dl = []

        planet_radius = self._planet.fullRadius
        total_layers = self.nLayers

        z = self.altitudeProfile
        self.debug('Computing path_length: \n z=%s \n dz=%s', z, dz)

        for layer in range(0, total_layers):

            p = (planet_radius+dz[0]/2 + z[layer])**2
            k = np.zeros(shape=(self.nLayers-layer))
            k[0] = np.sqrt((planet_radius + dz[0]/2. + z[layer] +
                            dz[layer]/2.)**2 - p)

            k[1:] = np.sqrt((planet_radius + dz[0]/2 + z[layer+1:] +
                             dz[layer+1:]/2)**2 - p)

            k[1:] -= np.sqrt((planet_radius + dz[0]/2 +
                              z[layer:self.nLayers-1] +
                              dz[layer:self.nLayers-1]/2)**2 - p)

            dl.append(k*2.0)
        return dl

    def path_integral(self, wngrid, return_contrib):

        dz = np.gradient(self.altitudeProfile)

        wngrid_size = wngrid.shape[0]

        path_length = self.compute_path_length(dz)

        density_profile = self.densityProfile

        total_layers = self.nLayers

        path_length = self.compute_path_length(dz)
        self.path_length = path_length

        tau = np.zeros(shape=(total_layers, wngrid_size), dtype=np.float64)

        for layer in range(total_layers):

            self.debug('Computing layer %s', layer)
            dl = path_length[layer]

            endK = total_layers-layer

            for contrib in self.contribution_list:
                if tau[layer].min() > 10:
                    break
                self.debug('Adding contribution from %s', contrib.name)
                contrib.contribute(self, 0, endK, layer, layer,
                                   density_profile, tau, path_length=dl)

        self.debug('tau %s %s', tau, tau.shape)

        absorption, tau = self.compute_absorption(tau, dz)
        return absorption, tau

    def compute_absorption(self, tau, dz):

        tau = np.exp(-tau)
        ap = self.altitudeProfile[:, None]
        pradius = self._planet.fullRadius
        sradius = self._star.radius
        _dz = dz[:, None]

        integral = np.sum((pradius+ap)*(1.0-tau)*_dz*2.0, axis=0)
        return ((pradius**2.0) + integral)/(sradius**2), tau



from taurex.data.spectrum.spectrum import BaseSpectrum
from taurex.data.spectrum.observed import ObservedSpectrum
from taurex.data.spectrum.array import ArraySpectrum
from taurex.data.spectrum.iraclis import IraclisSpectrum
try:
    from taurex.data.spectrum.lightcurve import ObservedLightCurve
except ImportError:
    print('pylightcurve not install. Ignoring')


from .cia import CIA
import pickle
from pathlib import Path
from taurex.util.math import interp_lin_only


class PickleCIA(CIA):
    """
    Class for using pickled (``.db``) collisionally induced absorptions
    Very simple since the format is simple


    Parameters
    ----------
    filename : str
        Path to pickle

    pair_name : str , optional
        Whilst the name of the pair is determined by the pickle filename
        since these can be different you can optionally force the name through 
        this parameter

    """

    def __init__(self, filename, pair_name=None):

        if pair_name is None:
            pair_name = Path(filename).stem

        super().__init__('PickleCIA', pair_name)

        self._filename = filename
        self._molecule_name = None
        self._spec_dict = None
        self._load_pickle_file(filename)

    def _load_pickle_file(self, filename):
        """
        Loads pickle file

        Parameters
        ----------
        filename : str
            Path to pickle cia file

        """

        # Load the pickle file
        self.info('Loading cia cross section from %s', filename)
        with open(filename, 'rb') as f:
            self._spec_dict = pickle.load(f, encoding='latin1')

        self._wavenumber_grid = self._spec_dict['wno']
        self._temperature_grid = self._spec_dict['t']
        self._xsec_grid = self._spec_dict['xsecarr']

    @property
    def wavenumberGrid(self):
        """

        Returns
        -------
        :obj:`array`
            Native wavenumber grid

        """
        return self._wavenumber_grid

    @property
    def temperatureGrid(self):
        """

        Returns
        -------
        :obj:`array`
            Native temperature grid in Kelvin

        """

        return self._temperature_grid

    def find_closest_temperature_index(self, temperature):
        """
        Finds the nearest indices for a particular temperature

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        Returns
        -------
        t_min : int
            index on temprature grid to the left of ``temperature``

        t_max : int
            index on temprature grid to the right of ``temperature``

        """
        t_min = self.temperatureGrid.searchsorted(temperature,
                                                  side='right')-1
        t_max = t_min+1
        return t_min, t_max

    def interp_linear_grid(self, T, t_idx_min, t_idx_max):
        """
        For a given temperature and indicies. Interpolate the cross-sections
        linearly from temperature grid to temperature ``T``

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        t_min : int
            index on temprature grid to the left of ``temperature``

        t_max : int
            index on temprature grid to the right of ``temperature``

        Returns
        -------
        out : :obj:`array`
            Interpolated cross-section

        """
        if T > self._temperature_grid.max():
            return self._xsec_grid[-1]
        elif T < self._temperature_grid.min():
            return self._xsec_grid[0]

        Tmax = self._temperature_grid[t_idx_max]
        Tmin = self._temperature_grid[t_idx_min]
        fx0 = self._xsec_grid[t_idx_min]
        fx1 = self._xsec_grid[t_idx_max]

        return interp_lin_only(fx0, fx1, T, Tmin, Tmax)

    def compute_cia(self, temperature):
        """
        Computes the collisionally induced absorption cross-section
        using our native temperature and cross-section grids

        Parameters
        ----------
        temperature : float
            Temperature in Kelvin

        Returns
        -------
        out : :obj:`array`
            Temperature interpolated cross-section

        """
        indicies = self.find_closest_temperature_index(temperature)
        return self.interp_linear_grid(temperature, *indicies)



"""Module contains classes that handle loading of HITRAN cia files"""
from .cia import CIA
import numpy as np
from taurex.util.math import interp_lin_only


class EndOfHitranCIAException(Exception):
    """
    An exception that occurs when the end of a HITRAN file
    is reached
    """
    pass


def hashwn(start_wn, end_wn):
    """
    Simple wavenumber hash function
    """
    return str(start_wn)+str(end_wn)


class HitranCiaGrid(object):
    """
    Class that handles a particular HITRAN cia wavenumber grid
    Since temperatures for CIA sometimes have different wavenumber grids this
    class helps to simplify managing them by only dealing with one at a time.
    These will help us unify into a single grid eventually

    Parameters
    ----------
    wn_min : float
        The minimum wavenumber for this grid

    wn_max : float
        The maximum wavenumber for this grid

    """

    def __init__(self, wn_min, wn_max):
        self.wn = None
        self.Tsigma = []

    def add_temperature(self, T, sigma):
        """
        Adds a temeprature and crossection to this wavenumber grid

        Parameters
        ----------
        T : float
            Temeprature in Kelvin

        sigma : :obj:`array`
            cross-sections for this grid

        """

        self.Tsigma.append((T, sigma))

    @property
    def temperature(self):
        """

        Gets the current temeprature grid for this wavenumber grid

        Returns
        -------
        :obj:`array`
            Temeprature grid in Kelvin

        """
        return [t for t, s in self.Tsigma]

    @property
    def sigma(self):
        """
        Gets the currently loaded crossections for this wavenumber grid

        Returns
        -------
        :obj:`array`
            Cross-section grid

        """
        return [s for t, s in self.Tsigma]

    def find_closest_temperature_index(self, temperature):
        """
        Finds the nearest indices for a particular temperature

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        Returns
        -------
        t_min : int
            index on temprature grid to the left of ``temperature``

        t_max : int
            index on temprature grid to the right of ``temperature``

        """
        t_min = np.array(self.temperature).searchsorted(temperature,
                                                        side='right')-1
        t_max = t_min+1
        return t_min, t_max

    def interp_linear_grid(self, T, t_idx_min, t_idx_max):
        """
        For a given temperature and indicies. Interpolate the cross-sections
        linearly from temperature grid to temperature ``T``

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin
        
        t_min : int
            index on temprature grid to the left of ``temperature``

        t_max : int
            index on temprature grid to the right of ``temperature``

        Returns
        -------
        out : :obj:`array`
            Interpolated cross-section

        """

        temp_grid = np.array(self.temperature)
        Tmax = temp_grid[t_idx_max]
        Tmin = temp_grid[t_idx_min]
        fx0 = self.sigma[t_idx_min]
        fx1 = self.sigma[t_idx_max]

        return interp_lin_only(fx0, fx1, T, Tmin, Tmax)

    def sortTempSigma(self):
        """
        Sorts the temperature-sigma list
        """
        import operator
        self.Tsigma.sort(key=operator.itemgetter(0))

    def fill_temperature(self, temperatures):
        """

        Here the 'master' temperature grid is passed into here and
        any gaps in our grid is filled with zero cross-sections to produce
        our final temperature-crosssection grid that matches with every other
        wavenumber grid. Temperatures that don't exist in the current grid but are withing
        the minimum and maximum for us are produced by linear interpolation


        Parameters
        ----------
        temperatures : array_like
            Master temperature grid

        """
        for t in temperatures:
            if t in self.temperature:
                continue
            self.debug('Tempurature %s, %s', t)
            if t < min(self.temperature) or t > max(self.temperature):
                self.add_temperature(t, np.zeros_like(self.wn))
            else:
                indicies = self.find_closest_temperature_index(t)
                self.add_temperature(t, self.interp_linear_grid(t, *indicies))
            self.sortTempSigma()


class HitranCIA(CIA):
    """

    A class that directly deals with HITRAN `cia <https://hitran.org/cia/>`_
    files and turns them into generic CIA objects that nicely produces
    cross sections for us.
    This will handle CIAs that have wavenumber grids split
    across temperatures by unifying
    them into single grids.

    To use it simply do:

    >>> h2h2 = HitranCIA('path/to/H2-He.cia')

    And now you can painlessly compute cross-sections like this:

    >>> h2h2.cia(400)

    Or if you have a wavenumber grid, we can also interpolate it:

    >>> h2h2.cia(400,mywngrid)

    And all it cost was buying me a beer!


    Parameters
    ----------
    filename : str
        Path to HITRAN cia file


    """

    def __init__(self, filename):
        super().__init__(self.__class__.__name__, 'None')

        self._filename = filename
        self._molecule_name = None
        self._wavenumber_grid = None
        self._temperature_grid = None
        self._xsec_grid = None
        self._wn_dict = {}
        self.load_hitran_file(filename)

    def load_hitran_file(self, filename):
        """
        Handles loading of the HITRAN file by reading and figuring
        out the wavenumber and temperature grids and matching them up

        Parameters
        ----------
        filename : str
            Path to HITRAN cia file

        """

        temp_list = []

        with open(filename, 'r') as f:
            # Read number of points
            while True:
                try:
                    start_wn, end_wn, total_points, T, \
                                max_cia = self.read_header(f)

                except EndOfHitranCIAException:
                    break
                if T not in temp_list:
                    temp_list.append(T)

                wn_hash = hashwn(start_wn, end_wn)

                wn_obj = None
                if wn_hash not in self._wn_dict:
                    self._wn_dict[wn_hash] = HitranCiaGrid(start_wn, end_wn)

                wn_obj = self._wn_dict[wn_hash]

                # Clear the temporary list
                sigma_temp = []
                wn_temp = []
                for x in range(total_points):
                    line = f.readline()
                    self.debug('Line %s', line)
                    splits = line.split()

                    _wn = splits[0]
                    _sigma = splits[1]
                    wn_temp.append(float(_wn))
                    _sig = float(_sigma)*1e-10
                    if _sig < 0:
                        _sig = 0
                    sigma_temp.append(_sig)

                # Ok we're done lets add the sigma
                wn_obj.add_temperature(T, np.array(sigma_temp))
                # set the wavenumber grid
                wn_obj.wn = np.array(wn_temp)

        temp_list.sort()
        self._temperature_grid = np.array(temp_list)
        self.fill_gaps(temp_list)
        self.compute_final_grid()

    def fill_gaps(self, temperature):
        """

        Fills gaps in temperature grid for all wavenumber grid objects
        we've created


        Parameters
        ----------
        temperature : array_like
            Master temperature grid


        """
        for wn_obj in self._wn_dict.values():
            wn_obj.sortTempSigma()
            wn_obj.fill_temperature(temperature)

    def compute_final_grid(self):
        """

        Collects all :class:`HitranCiaGrid` objects we've created
        and unifies them into a single temperature, cross-section and
        wavenumber grid for us to FINALLY interpolate and produce
        collisionaly induced cross-sections

        """

        _wngrid = []
        for w in self._wn_dict.values():
            _wngrid.append(w.wn)
        self._wavenumber_grid = np.concatenate(_wngrid)
        sorted_idx = np.argsort(self._wavenumber_grid)
        self._wavenumber_grid = self._wavenumber_grid[sorted_idx]
        _sigma_array = []

        for idx, t in enumerate(self._temperature_grid):
            _temp_sigma = []
            for w in self._wn_dict.values():
                _temp_sigma.append(w.Tsigma[idx][1])
            _sigma_array.append(np.concatenate(_temp_sigma)[sorted_idx])

        self._xsec_grid = np.array(_sigma_array)

    def find_closest_temperature_index(self, temperature):
        """
        Finds the nearest indices for a particular temperature

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        Returns
        -------
        t_min : int
            index on temprature grid to the left of ``temperature``

        t_max : int
            index on temprature grid to the right of ``temperature``

        """
        t_min = self.temperatureGrid.searchsorted(temperature, side='right')-1
        t_max = t_min+1
        return t_min, t_max

    def interp_linear_grid(self, T, t_idx_min, t_idx_max):
        """
        For a given temperature and indicies. Interpolate the cross-sections
        linearly from temperature grid to temperature ``T``

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        t_min : int
            index on temprature grid to the left of ``temperature``

        t_max : int
            index on temprature grid to the right of ``temperature``

        Returns
        -------
        out : :obj:`array`
            Interpolated cross-section

        """
        if T > self._temperature_grid.max():
            return self._xsec_grid[-1]
        elif T < self._temperature_grid.min():
            return self._xsec_grid[0]

        Tmax = self._temperature_grid[t_idx_max]
        Tmin = self._temperature_grid[t_idx_min]
        fx0 = self._xsec_grid[t_idx_min]
        fx1 = self._xsec_grid[t_idx_max]

        return interp_lin_only(fx0, fx1, T, Tmin, Tmax)


    def read_header(self, f):
        """
        Reads single header in the file

        Parameters
        ----------
        f : file object

        Returns
        -------
        start_wn : float
            Start wavenumber for temperature

        end_wn : float
            End wavenumber for temperature

        total_points : int
            total number of points in temperature

        T : float
            Temperature in Kelvin

        max_cia : float
            Maximum CIA value in temperature


        """

        line = f.readline()
        if line is None or line == '':
            raise EndOfHitranCIAException
        split = line.split()
        self._pair_name = split[0]
        start_wn = float(split[1])
        end_wn = float(split[2])
        total_points = int(split[3])
        T = float(split[4])
        max_cia = float(split[5])

        return start_wn, end_wn, total_points, T, max_cia

    @property
    def wavenumberGrid(self):
        """
        Unified wavenumber grid

        Returns
        -------
        :obj:`array`
            Native wavenumber grid

        """

        return self._wavenumber_grid

    @property
    def temperatureGrid(self):
        """
        Unified temperature grid

        Returns
        -------
        :obj:`array`
            Native temperature grid in Kelvin

        """
        return self._temperature_grid

    def compute_cia(self, temperature):
        """
        Computes the collisionally induced absorption cross-section
        using our final native temperature and cross-section grids

        Parameters
        ----------
        temperature : float
            Temperature in Kelvin

        Returns
        -------
        out : :obj:`array`
            Temperature interpolated cross-section

        """

        indicies = self.find_closest_temperature_index(temperature)

        return self.interp_linear_grid(temperature, *indicies)



"""Module for the base binning class"""

from taurex.log import Logger
from taurex import OutputSize
from taurex.util.util import compute_bin_edges


class Binner(Logger):
    """
    *Abstract class*

    The binner class deals with binning down spectra to different resolutions.
    It also provides a method to generate spectrum output format from a forward
    model result in the form of a dictionary.
    Using this class does not need to be restricted to TauREx3 results and can
    be used to bin down any arbitrary spectra.
    """
    def __init__(self):
        Logger.__init__(self, self.__class__.__name__)

    def bindown(self, wngrid, spectrum, grid_width=None, error=None):
        """
        **Requires implementation**

        This should handle the binning of a spectrum passed into the function.
        Parameters given are guidelines on expectation of usage.

        Parameters
        ----------
        wngrid : :obj:`array`
            The wavenumber grid of the spectrum to be binned down.
            Generally the 'native' wavenumber grid

        spectrum: :obj:`array`
            The spectra we wish to bin-down. Must be same shape as
            ``wngrid``.

        grid_width: :obj:`array`, optional
            Wavenumber grid full-widths for the spectrum to be binned down.
            Must be same shape as ``wngrid``.
            Optional, generally if you require this but the user does not pass
            it then you must compute it yourself using ``wngrid``. This can
            be done easily using the function
            func:`~taurex.util.util.compute_bin_edges`.

        error: :obj:`array`, optional
            Associated errors or noise of the spectrum. Must be same shape
            as ``wngrid``.Optional parameter, when implementing you must
            deal with the cases where either the error is passed or not passed.

        Returns
        -------
        binned_wngrid : :obj:`array`
            New wavenumber grid

        spectrum: :obj:`array`
            Binned spectrum.

        grid_width: :obj:`array`
            New grid-widths

        error: :obj:`array` or None
            If passed, should be the binned error otherwise None

        """
        raise NotImplementedError

    def bin_model(self, model_output):
        """
        Bins down a TauREx3 forward model.
        This automatically splits the output and passes it to
        the :func:`bindown` function.
        Its general usage is of the form:

        >>> fm = TransmissionModel()
        >>> fm.build()
        >>> result = fm.model()
        >>> binner.bin_model(result)

        Or in a single line:

        >>> binner.bin_model(fm.model())


        Parameters
        ----------
        model_output: obj:`tuple`
            Result from running a TauREx3 forward model
        
        Returns
        -------
        See :func:`bindown`


        """
        return self.bindown(model_output[0], model_output[1])

    def generate_spectrum_output(self, model_output,
                                 output_size=OutputSize.heavy):
        """
        Given a forward model output, generate a dictionary
        that can be used to store to file. This can include
        storing the native and binned spectrum.
        Not necessary for the function of the class but useful for
        full intergation into TauREx3, especially when storing results
        from a retrieval. 
        Can be overwritten to store more information.

        Parameters
        ----------
        model_output: obj:`tuple`
            Result from running a TauREx3 forward model

        output_size: :class:`~taurex.taurexdefs.OutputSize`
            Size of the output.


        Returns
        -------
        :obj:`dict`:
            Dictionary of spectra


        """
        output = {}

        wngrid, flux, tau, extra = model_output

        output['native_wngrid'] = wngrid
        output['native_wlgrid'] = 10000/wngrid
        output['native_spectrum'] = flux
        output['binned_spectrum'] = self.bindown(wngrid, flux)[1]
        output['native_wnwidth'] = compute_bin_edges(wngrid)[-1]
        output['native_wlwidth'] = compute_bin_edges(10000/wngrid)[-1]
        if output_size > OutputSize.lighter:
            output['binned_tau'] = self.bindown(wngrid, tau)[1]
            if output_size > OutputSize.light:
                output['native_tau'] = tau

        return output




from .output import Output, OutputGroup
import h5py
from taurex.mpi import get_rank, only_master_rank
import datetime


class HDF5OutputGroup(OutputGroup):

    def __init__(self, entry):
        super().__init__('HDF5Group')
        self._entry = entry

    @only_master_rank
    def write_array(self, array_name, array, metadata=None):
        ds = self._entry.create_dataset(
            str(array_name), data=array, shape=array.shape, dtype=array.dtype)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    @only_master_rank
    def write_scalar(self, scalar_name, scalar, metadata=None):
        ds = self._entry.create_dataset(str(scalar_name), data=scalar)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    @only_master_rank
    def write_string(self, string_name, string, metadata=None):
        ds = self._entry.create_dataset(str(string_name), data=string)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def create_group(self, group_name):
        entry = None
        if self._entry:
            entry = self._entry.create_group(str(group_name))
        return HDF5OutputGroup(entry)

    @only_master_rank
    def write_string_array(self, string_name, string_array, metadata=None):

        asciiList = [n.encode("ascii", "ignore") for n in string_array]
        ds = self._entry.create_dataset(
            str(string_name), (len(asciiList), 1), 'S64', asciiList)

        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v


class HDF5Output(Output):
    def __init__(self, filename, append=False):
        super().__init__('HDF5Output')
        self.filename = filename
        self._append = append
        self.fd = None

    def open(self):
        if get_rank() == 0:
            self.fd = self._openFile(self.filename, self._append)

    def _openFile(self, fname, append):

        mode = 'w'
        if self._append:
            mode = 'a'

        fd = h5py.File(fname, mode=mode)
        fd.attrs['file_name'] = fname
        fd.attrs['file_time'] = datetime.datetime.now().isoformat()
        fd.attrs['creator'] = self.__class__.__name__
        fd.attrs['HDF5_Version'] = h5py.version.hdf5_version
        fd.attrs['h5py_version'] = h5py.version.version
        fd.attrs['program_name'] = 'TauREx'
        fd.attrs['program_version'] = 'v3.0'
        return fd

    def create_group(self, group_name):
        entry = None
        if self.fd:
            entry = self.fd.create_group(str(group_name))
        return HDF5OutputGroup(entry)

    def close(self):
        if self.fd:
            self.fd.flush()
            self.fd.close()



from taurex.model import ForwardModel
import numpy as np

import pickle
import pylightcurve as plc
from .lightcurvedata import LightCurveData
from taurex.data.fittable import fitparam


class LightCurveModel(ForwardModel):
    """A base class for producing forward models"""

    @property
    def defaultBinner(self):
        from taurex.binning.lightcurvebinner import LightcurveBinner
        return LightcurveBinner

    def __init__(self, forward_model, file_loc, instruments=None):
        super().__init__('LightCurveModel')

        self._forward_model = forward_model
        self.file_loc = file_loc
        self._load_file()

        self._load_orbital_profile()

        if instruments is None:
            instruments = 'all'

        if instruments == 'all' or 'all' in instruments:
            instruments = LightCurveData.availableInstruments

        self._load_instruments(instruments)
        self._load_ldcoeff()
        self._initialize_lightcurves()
        self.setup_binner()

    def setup_binner(self):
        from taurex.data.spectrum import ArraySpectrum
        self._wngrid = 10000 / self.lc_data['obs_spectrum'][:, 0]
        self._arr_spec = ArraySpectrum(self.lc_data['obs_spectrum'])
        self._binner = self._arr_spec.create_binner()

    def initialize_profiles(self):
        self._forward_model.initialize_profiles()

    def _load_file(self):
        # input data from lightcurve, not from spectrum.

        with open(self.file_loc, 'rb') as f:
            self.lc_data = pickle.load(f, encoding='latin1')

    def _load_orbital_profile(self):
        """Load orbital information"""
        self.info('Load lc information')

        # new
        self.mid_time = self.lc_data['orbital_info']['mt']
        self._inclination = self.lc_data['orbital_info']['i']
        self.period = self.lc_data['orbital_info']['period']
        self.periastron = self.lc_data['orbital_info']['periastron']
        self.sma_over_rs = self.lc_data['orbital_info']['sma_over_rs']
        self.ecc = self.lc_data['orbital_info']['e']

    def _load_ldcoeff(self):
        # new

        self.ld_coeff_file = np.array([])
        for i in self._instruments:
            self.ld_coeff_file = np.append(
                self.ld_coeff_file,
                self.lc_data[i.instrumentName]['ld_coeff']).reshape(-1, 4)
        assert np.shape(self.ld_coeff_file)[
            1] == 4, "please use 4 ldcoeff law."

    def _load_instruments(self, instruments):
        self._instruments = []
        # new
        ins_keys = self.lc_data.keys()
        for ins in instruments:
            if ins in ins_keys:
                self.info('Loading {} light curves'.format(ins))
                self._instruments.append(
                    LightCurveData.fromInstrumentName(ins, self.lc_data))
            else:
                self.info('Could not find {} in instrument keys'.format(ins))
        new_instrument_list = sorted(
            self._instruments, key=lambda x: x.wavelengthRegion[0],
            reverse=True)
        self._instruments = new_instrument_list

    def _initialize_lightcurves(self):

        nFactor = [ins.minNFactors for ins in self._instruments]

        min_n_factors = None

        max_n_factors = None
        if len(nFactor) > 0:
            min_n_factors = np.concatenate(nFactor)

        nFactor = [ins.maxNFactors for ins in self._instruments]
        if len(nFactor) > 0:
            max_n_factors = np.concatenate(nFactor)

        if max_n_factors is not None and min_n_factors is not None:
            self.create_normalization_fitparams()

            for idx, value in enumerate(zip(min_n_factors, max_n_factors)):
                min_n, max_n = value
                self.modify_bounds('Nfactor_{}'.format(idx), [min_n, max_n])

        min_time = min([ins.timeSeries.min() for ins in self._instruments])

        max_time = max([ins.timeSeries.max() for ins in self._instruments])

        self.modify_bounds('mid_transit_time', [min_time, max_time])

    @fitparam(param_name='sma_over_rs', param_latex='sma_over_rs',
              default_mode='linear', default_fit=False,
              default_bounds=[1.001, 60.0])
    def semiMajorAxisOverRs(self):
        return self.sma_over_rs

    @semiMajorAxisOverRs.setter
    def semiMajorAxisOverRs(self, value):
        self.sma_over_rs = value

    @fitparam(param_name='mid_transit_time', param_latex='mid_transit_time',
              default_mode='linear', default_fit=False,
              default_bounds=[1.001, 2.0])
    def midTransitTime(self):
        return self.mid_time

    @midTransitTime.setter
    def midTransitTime(self, value):
        self.mid_time = value

    @fitparam(param_name='inclination', param_latex='inclination',
              default_mode='linear', default_fit=False,
              default_bounds=[40.0, 90.0])
    def inclination(self):
        return self._inclination

    @inclination.setter
    def inclination(self, value):
        self._inclination = value

    @property
    def temperatureProfile(self):
        return self._forward_model.temperatureProfile

    @property
    def pressureProfile(self):
        return self._forward_model.pressureProfile

    @property
    def densityProfile(self):
        return self._forward_model.densityProfile

    @property
    def scaleheight_profile(self):
        return self._forward_model.scaleheight_profile

    @property
    def chemistry(self):
        return self._forward_model.chemistry

    @property
    def gravity_profile(self):
        return self._forward_model.gravity_profile

    @property
    def pressure(self):
        return self._forward_model.pressure

    @property
    def altitudeProfile(self):
        return self._forward_model.altitudeProfile

    def create_normalization_fitparams(self):
        import itertools

        ins_name = list(itertools.chain(
            *tuple([[ins.instrumentName]*ins.minNFactors.shape[0]
                    for ins in self._instruments])))

        ins_number = list(itertools.chain(
            *tuple([list(range(ins.minNFactors.shape[0]))
                    for ins in self._instruments])))

        for idx, val in enumerate(zip(ins_name, ins_number)):
            name, no = val
            param_name = 'Nfactor_{}'.format(idx)
            param_latex = '{}_{}'.format(name, no)

            default_fit = False
            default_bounds = [0, 1]
            default_mode = 'linear'

            def readN(self, idx=idx):
                return self._nfactor[idx]

            def writeN(self, value, idx=idx):
                self._nfactor[idx] = value

            self.add_fittable_param(
                param_name, param_latex, readN, writeN, default_mode,
                default_fit, default_bounds)

    def instrument_light_curve(self, model, wlgrid):
        """Combine light-curves from different instrucments together."""
        result = np.array([])
        sma_over_rs_value = self.sma_over_rs
        inclination_value = self.inclination
        mid_time_value = self.mid_time
        try:
            Nfactor = self._nfactor
        except AttributeError:
            Nfactor = np.ones_like(wlgrid)

        result = []

        for ins in self._instruments:
            min_wl, max_wl = ins.wavelengthRegion
            index = (wlgrid > min_wl) & (wlgrid < max_wl)

            lc = self.light_curve_chain(model[index],
                                        time_array=ins.timeSeries,
                                        period=self.period,
                                        sma_over_rs=sma_over_rs_value,
                                        eccentricity=self.ecc,
                                        inclination=inclination_value,
                                        periastron=self.periastron,
                                        mid_time=mid_time_value,
                                        ldcoeff=self.ld_coeff_file[index],
                                        Nfactor=Nfactor[index])
            result.append(lc)

        return np.concatenate(result)

    def light_curve_chain(self, model, time_array, period, sma_over_rs,
                          eccentricity, inclination, periastron,
                          mid_time, ldcoeff, Nfactor):
        """Create model light-curve and lightcurve chain."""
        from taurex.model import TransmissionModel, EmissionModel
        result = []
        sqrt_model = np.sqrt(model)
        # self.info('Creating Lightcurve chain.')
        for n in range(len(sqrt_model)):
            if isinstance(self._forward_model, TransmissionModel):
                self.debug('Using Transit')
                transit_light_curve = plc.transit('claret', ldcoeff[n],
                                                  sqrt_model[n], period,
                                                  sma_over_rs, eccentricity,
                                                  inclination, periastron,
                                                  mid_time, time_array)
                result.append(transit_light_curve * Nfactor[n])
            elif isinstance(self._forward_model, EmissionModel):
                self.debug('Using Eclipse')
                rp_over_rs = (self._forward_model.planet.fullRadius /
                              self._forward_model.star.radius)
                self.debug('rp_over_rs %s', rp_over_rs)
                self.debug('fp_over_fs %s', sqrt_model)

                eclipse_light_curve = plc.eclipse(sqrt_model[n], rp_over_rs,
                                                  period,
                                                  sma_over_rs, eccentricity,
                                                  inclination, periastron,
                                                  mid_time, time_array)
                result.append(eclipse_light_curve * Nfactor[n])

        self.debug('Result %s', result)
        return np.concatenate(result)

    def build(self):
        self._fitting_parameters = {}
        self._fitting_parameters.update(self.fitting_parameters())

        self._forward_model.build()

        self._fitting_parameters.update(self._forward_model.fittingParameters)

    @property
    def nativeWavenumberGrid(self):
        return self._forward_model.nativeWavenumberGrid

    def model(self, wngrid=None, cutoff_grid=True):
        """Computes the forward model for a wngrid"""
        # new

        native_grid, model, tau, extra = self._forward_model.model(
            self._wngrid, cutoff_grid)
        binned_model = self._binner.bindown(native_grid, model)
        wlgrid = 10000/self._wngrid
        result = self.instrument_light_curve(binned_model[1], wlgrid)

        return self._wngrid, result, tau, \
            [native_grid, model, binned_model[1], extra]

    def compute_error(self, samples, wngrid=None, binner=None):
        from taurex.util.math import OnlineVariance
        tp_profiles = OnlineVariance()
        active_gases = OnlineVariance()
        inactive_gases = OnlineVariance()

        lc_spectrum = OnlineVariance()

        native_spectrum = OnlineVariance()
        binned_spectrum = OnlineVariance()

        for weight in samples():

            grid, lc, tau, extra = self.model(wngrid=wngrid, cutoff_grid=False)
            native_grid, native, binned, _ = extra

            tp_profiles.update(self.temperatureProfile, weight=weight)
            active_gases.update(
                self.chemistry.activeGasMixProfile, weight=weight)
            inactive_gases.update(
                self.chemistry.inactiveGasMixProfile, weight=weight)

            native_spectrum.update(native, weight=weight)
            binned_spectrum.update(binned, weight=weight)
            lc_spectrum.update(lc, weight=weight)

        profile_dict = {}
        spectrum_dict = {}

        tp_std = np.sqrt(tp_profiles.parallelVariance())
        active_std = np.sqrt(active_gases.parallelVariance())
        inactive_std = np.sqrt(inactive_gases.parallelVariance())

        profile_dict['temp_profile_std'] = tp_std
        profile_dict['active_mix_profile_std'] = active_std
        profile_dict['inactive_mix_profile_std'] = inactive_std

        spectrum_dict['native_std'] = np.sqrt(
            native_spectrum.parallelVariance())
        spectrum_dict['binned_std'] = np.sqrt(
            binned_spectrum.parallelVariance())
        spectrum_dict['lightcurve_std'] = np.sqrt(
            lc_spectrum.parallelVariance())
        return profile_dict, spectrum_dict

    def model_contrib(self, wngrid=None, cutoff_grid=True):
        # new
        wngrid = 10000 / self.lc_data['obs_spectrum'][:, 0]
        native_grid, contribs = self._forward_model.model_contrib(
            wngrid, cutoff_grid)
        binner = self._binner
        all_contrib_dict = {}

        wlgrid = 10000/wngrid

        for contrib_name, main_contribution in contribs.items():

            model, tau, extras = main_contribution
            binned = binner.bindown(native_grid, model)[1]

            result = self.instrument_light_curve(binned, wlgrid)
            all_contrib_dict[contrib_name] = (
                result, tau, [native_grid, model, binned, extras])

        return native_grid, all_contrib_dict

    def model_full_contrib(self, wngrid=None, cutoff_grid=True):
        """Computes the forward model for a wngrid for each contribution"""

        # new
        wngrid = 10000 / self.lc_data['obs_spectrum'][:, 0]

        native_grid, contrib_res = self._forward_model.model_full_contrib(
            wngrid, cutoff_grid)
        binner = self._binner

        self.info('Computing lightcurve contribution')
        wlgrid = 10000/wngrid
        lc_contrib_res = {}

        # Loop through each contribtuion
        for contrib_name, contrib_list in contrib_res.items():

            lc_contrib_list = []

            for name, native, tau, extra in contrib_list:

                binned = binner.bindown(native_grid, native)[1]
                result = self.instrument_light_curve(binned, wlgrid)

                new_packed = name, result, tau, (native_grid,
                                                 native, binned, extra)

                lc_contrib_list.append(new_packed)

            lc_contrib_res[contrib_name] = lc_contrib_list

        return native_grid, lc_contrib_res

    def write(self, output):
        lc = output.create_group('Lightcurve')

        lc_grps = lc.create_group('Instrument')
        for ins in self._instruments:
            ins.write(lc_grps)

        lc.write_scalar('mid_time', self.mid_time)
        lc.write_scalar('inclination', self._inclination)
        lc.write_scalar('period', self.period)
        lc.write_scalar('periastron', self.periastron)
        lc.write_scalar('sma_over_rs', self.sma_over_rs)
        lc.write_scalar('eccentricity', self.ecc)

        self._forward_model.write(output)

        return None



import pathlib
from taurex.opacity.interpolateopacity import InterpolatingOpacity
import numpy as np


class ExoTransmitOpacity(InterpolatingOpacity):

    def __init__(self, filename, interpolation_mode='linear', in_memory=False):

        super().__init__('ExoOpacity:{}'.format(
                            pathlib.Path(filename).stem[4:]),
                         interpolation_mode=interpolation_mode)

        self._filename = filename
        self._molecule_name = pathlib.Path(filename).stem[4:]
        self.in_memory = in_memory
        self._load_exo_transmit(filename)

    def _load_exo_transmit(self, filename):
        self.debug('Loading opacity from {}'.format(filename))

        with open(filename, 'r') as f:
            lines = f.readlines()

        self._temperature_grid = np.array(
            [float(l) for l in lines[0].split()])  # *t_conversion

        self._pressure_grid = np.array(
            [float(l) for l in lines[1].split()])*1e5

        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()

        wn_grid = []

        for ln in lines[2:]:
            arr = np.array([float(l) for l in ln.split()])
            if arr.shape[0] == 1:
                wn_grid.append(10000*1e-6/arr[0])

        wn_grid = np.array(wn_grid)

        grid_sort = wn_grid.argsort()
        self._wavenumber_grid = wn_grid[grid_sort]

        pressure_count = 0
        lambda_count = -1
        self._xsec_grid = np.empty(shape=(self.pressureGrid.shape[0],
                                          self.temperatureGrid.shape[0],
                                          self.wavenumberGrid.shape[0]))

        for ln in lines[2:]:
            arr = np.array([float(l) for l in ln.split()])
            if arr.shape[0] == 1:
                lambda_count += 1
                pressure_count = 0
            else:
                self._xsec_grid[pressure_count, :,
                                lambda_count] = arr[1:] + 1e-60
                pressure_count += 1

        self._xsec_grid = self._xsec_grid[:, :, grid_sort]*10000

    @property
    def wavenumberGrid(self):
        return self._wavenumber_grid

    @property
    def temperatureGrid(self):
        return self._temperature_grid

    @property
    def pressureGrid(self):
        return self._pressure_grid

    @property
    def resolution(self):
        return self._resolution

    @property
    def moleculeName(self):
        return self._molecule_name

    @property
    def xsecGrid(self):
        return self._xsec_grid



from .logger import Logger
import logging

last_log = logging.INFO


def setLogLevel(level):
    global last_log
    from .logger import root_logger
    root_logger.setLevel(level)
    last_log = level


def disableLogging():
    import logging
    from .logger import root_logger
    root_logger.setLevel(logging.ERROR)


def enableLogging():
    global last_log
    import logging
    if last_log is None:
        last_log = logging.INFO
    setLogLevel(last_log)



from .contribution import Contribution
import numpy as np
from taurex.data.fittable import fitparam


class SimpleCloudsContribution(Contribution):
    """
    Optically thick cloud deck up to a certain height

    These have the form:

    .. math::
            \\tau(\\lambda,z) =
                \\begin{cases}
                \\infty       & \\quad \\text{if } P(z) >= P_{0}\\\\
                0            & \\quad \\text{if } P(z) < P_{0}
                \\end{cases}

    Where :math:`P_{0}` is the pressure at the top of the cloud-deck


    Parameters
    ----------
    clouds_pressure: float
        Pressure at top of cloud deck


    """
    def __init__(self, clouds_pressure=1e3):
        super().__init__('SimpleClouds')
        self._cloud_pressure = clouds_pressure

    @property
    def order(self):
        return 3

    def contribute(self, model, start_layer, end_layer, density_offset, layer,
                   density, tau, path_length=None):
        tau[layer] += self.sigma_xsec[layer, :]

    def prepare_each(self, model, wngrid):
        """
        Returns an absorbing cross-section that is infinitely absorping
        up to a certain height

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            ``Clouds`` and opacity array.


        """

        contrib = np.zeros(shape=(model.nLayers, wngrid.shape[0],))
        cloud_filtr = model.pressureProfile >= self._cloud_pressure
        contrib[cloud_filtr, :] = np.inf
        self._contrib = contrib
        yield 'Clouds', self._contrib

    @fitparam(param_name='clouds_pressure',
              param_latex='$P_\mathrm{clouds}$',
              default_mode='log',
              default_fit=False, default_bounds=[1e-3, 1e6])
    def cloudsPressure(self):
        """
        Cloud top pressure in Pascal
        """
        return self._cloud_pressure

    @cloudsPressure.setter
    def cloudsPressure(self, value):
        self._cloud_pressure = value

    def write(self, output):
        contrib = super().write(output)
        contrib.write_scalar('clouds_pressure', self._cloud_pressure)
        return contrib



import numpy as np
from .simplemodel import SimpleForwardModel
from taurex.constants import PI
from taurex.util.emission import black_body


class EmissionModel(SimpleForwardModel):
    """
    A forward model for eclipses

    Parameters
    ----------

    planet: :class:`~taurex.data.planet.Planet`, optional
        Planet model, default planet is Jupiter

    star: :class:`~taurex.data.stellar.star.Star`, optional
        Star model, default star is Sun-like

    pressure_profile: :class:`~taurex.data.profiles.pressure.pressureprofile.PressureProfile`, optional
        Pressure model, alternative is to set ``nlayers``, ``atm_min_pressure``
        and ``atm_max_pressure``

    temperature_profile: :class:`~taurex.data.profiles.temperature.tprofile.TemperatureProfile`, optional
        Temperature model, default is an :class:`~taurex.data.profiles.temperature.isothermal.Isothermal`
        profile at 1500 K

    chemistry: :class:`~taurex.data.profiles.chemistry.chemistry.Chemistry`, optional
        Chemistry model, default is
        :class:`~taurex.data.profiles.chemistry.taurexchemistry.TaurexChemistry` with
        ``H2O`` and ``CH4``

    nlayers: int, optional
        Number of layers. Used if ``pressure_profile`` is not defined.

    atm_min_pressure: float, optional
        Pressure at TOA. Used if ``pressure_profile`` is not defined.

    atm_max_pressure: float, optional
        Pressure at BOA. Used if ``pressure_profile`` is not defined.

    ngauss: int, optional
        Number of gaussian quadrature points, default = 4

    """

    def __init__(self,
                 planet=None,
                 star=None,
                 pressure_profile=None,
                 temperature_profile=None,
                 chemistry=None,
                 nlayers=100,
                 atm_min_pressure=1e-4,
                 atm_max_pressure=1e6,
                 ngauss=4,
                 ):
        super().__init__(self.__class__.__name__,
                         planet,
                         star,
                         pressure_profile,
                         temperature_profile,
                         chemistry,
                         nlayers,
                         atm_min_pressure,
                         atm_max_pressure)

        self.set_num_gauss(ngauss)

    def set_num_gauss(self, value):
        self._ngauss = int(value)

        mu, weight = np.polynomial.legendre.leggauss(self._ngauss*2)
        self._mu_quads = mu[self._ngauss:]
        self._wi_quads = weight[self._ngauss:]

    def compute_final_flux(self, f_total):
        star_sed = self._star.spectralEmissionDensity

        self.debug('Star SED: %s', star_sed)
        # quit()
        star_radius = self._star.radius
        planet_radius = self._planet.fullRadius
        self.debug('star_radius %s', self._star.radius)
        self.debug('planet_radius %s', self._star.radius)
        last_flux = (f_total/star_sed) * (planet_radius/star_radius)**2

        self.debug('last_flux %s', last_flux)

        return last_flux

    def path_integral(self, wngrid, return_contrib):
        dz = np.gradient(self.altitudeProfile)

        density = self.densityProfile

        wngrid_size = wngrid.shape[0]

        total_layers = self.nLayers

        temperature = self.temperatureProfile
        tau = np.zeros(shape=(self.nLayers, wngrid_size))
        surface_tau = np.zeros(shape=(1, wngrid_size))

        layer_tau = np.zeros(shape=(1, wngrid_size))

        dtau = np.zeros(shape=(1, wngrid_size))

        # Do surface first
        # for layer in range(total_layers):
        for contrib in self.contribution_list:
            contrib.contribute(self, 0, total_layers, 0, 0,
                               density, surface_tau, path_length=dz)
        self.debug('density = %s', density[0])
        self.debug('surface_tau = %s', surface_tau)

        BB = black_body(wngrid, temperature[0])/PI

        _mu = 1.0/self._mu_quads[:, None]
        _w = self._wi_quads[:, None]
        I = BB * (np.exp(-surface_tau*_mu))

        self.debug('I1_pre %s', I)
        # Loop upwards
        for layer in range(total_layers):
            layer_tau[...] = 0.0
            dtau[...] = 0.0
            for contrib in self.contribution_list:
                contrib.contribute(self, layer+1, total_layers,
                                   0, 0, density, layer_tau, path_length=dz)
                contrib.contribute(self, layer, layer+1, 0,
                                   0, density, dtau, path_length=dz)

            _tau = np.exp(-layer_tau) - np.exp(-dtau)

            tau[layer] += _tau[0]
            # for contrib in self.contribution_list:

            self.debug('Layer_tau[%s]=%s', layer, layer_tau)

            dtau += layer_tau
            self.debug('dtau[%s]=%s', layer, dtau)
            BB = black_body(wngrid, temperature[layer])/PI
            self.debug('BB[%s]=%s,%s', layer, temperature[layer], BB)
            I += BB * (np.exp(-layer_tau*_mu) - np.exp(-dtau*_mu))

        self.debug('I: %s', I)

        flux_total = 2.0 * np.pi * sum(I * (_w / _mu))
        self.debug('flux_total %s', flux_total)

        return self.compute_final_flux(flux_total).flatten(), tau

    def write(self, output):
        model = super().write(output)
        model.write_scalar('ngauss', self._ngauss)
        return model




"""
This module relates to defining fitting parameters in TauREx3
"""

from functools import partial


def fitparam(f=None, param_name=None, param_latex=None,
             default_mode='linear', default_fit=False, 
             default_bounds=[0.0, 1.0]):
    """
    A decorator used in conjunction with :class:`Fittable` to inform which
    parameters can be fit and its properties. On its own it acts like the
    ``property`` decorator. When used within a :class:`Fittable` class it
    serves to tag a property as able to fit and allows the
    class to compile all parameters that can be fit.

    Its usage is simple, simply wrap a method and define its properties:

    .. code-block:: python

        class Foo(Fittable):

            @fitparam(param_name='foobar',param_latex='$Foo^{bar}$')
            def bar(self):
                return 'Foobar'

            @bar.setter
            def bar(self,value):
                self.value = value


    Parameters
    ----------
    f: function
        Function being passed. Automatically done when used as a decorator

    param_name: str
        Nicer name of the parameter. Referenced by the optimizer.

    param_latex: str
        Latex version of the parameter name, useful in plotting and
        making figures

    default_mode: ``linear`` or ``log``
        Defines how the optimizer should read and write the parameter.
        ``linear`` reads/write everything as is.
        ``log`` informs the optimizer to transform from native->log space
        when read and to transfrom log->native when
        writing. This also applies to the boundaries

    default_fit: bool
        Whether this is included in the fit without the user explicity saying
        so (Default: False)

    default_bounds: :obj:`list`
        Default minimum and maximum fitting boundary. Must always be defined
        in linear space


    """
    if f is None:
        return partial(fitparam, param_name=param_name,
                       param_latex=param_latex,
                       default_mode=default_mode,
                       default_fit=default_fit,
                       default_bounds=default_bounds)

    def wrapper(self, *args, **kwargs):

        return f(self, *args, **kwargs)
    wrapper.param_name = param_name
    wrapper.param_latex = param_latex
    wrapper.default_fit = default_fit
    wrapper.default_bounds = default_bounds
    wrapper.default_mode = default_mode
    wrapper.decorated = 'fitparam'
    pwrap = property(wrapper)

    return pwrap


class Fittable(object):
    """

    A class that manages fitting parameters. Not really used on its own
    it should really be inherited from to be used properly. It also provides
    class with the ability to read and write fitting parameters using their 
    params names, for example, if we create a class like this:

    .. code-block:: python

        class Foo(Fittable):

            def __init__(self):
                self.value = 10

            @fitparam(param_name='foobar',param_latex='$Foo^{bar}$',default_bounds=[1,12])
            def bar(self):
                return self.value

            @bar.setter
            def bar(self,value):
                self.value = value

    We can read and write data in the standard python way like so:

    >>> foo = Foo()
    >>> foo.bar
    10
    >>> foo.bar = 20
    >>> foo.bar
    20

    but we also get this functionality for free:

    >>> foo['foobar']
    20
    >>> foo['foobar'] = 30
    >>> foo['foobar']
    30


    """

    def __init__(self):
        self._param_dict = {}

        self.compile_fitparams()

    def add_fittable_param(self, param_name, param_latex, fget, fset,
                           default_mode, default_fit, default_bounds):
        """
        Adds a fittable parameter to the internal dictionary. Used during init
        to add all :func:`fitparam` decorated methods and can also be utilized
        by a user to manually add new fitting parameters. This is
        useful for giving fitting parameters names that depend on certain
        attributes (e.g. molecule name in a gas profile
        see
        :class:`~taurex.data.profiles.chemistry.gas.constantgas.ConstantGas`)
        or when converting lists into fitting parameters
        (e.g. Normalization factor in light curves
        see:
        :class:`~taurex.model.lightcurve.lightcurve.LightCurveModel` )

        Parameters
        ----------
        param_name: str
            Nicer name of the parameter. Referenced by the optimizer.

        param_latex: str
            Latex version of the parameter name, useful in plotting and making
            figures

        fget: function
            a function that returns the value of the parameter

        fset: function
            a function the writes the value of the parameter

        default_mode: ``linear`` or ``log``
            Defines how the optimizer should read and write the parameter.
            ``linear`` reads/write everything as is.
            ``log`` informs the optimizer to transform from native->log space
            when read and to transfrom log->native when writing. This also
            applies to the boundaries

        default_fit: bool
            Whether this is included in the fit without the user explicity
            saying so (Default: False)

        default_bounds: :obj:`list`
            Default minimum and maximum fitting boundary. Must always be
            defined in the native space

        """
        if param_name in self._param_dict:
            raise AttributeError(
                'param name {} already exists'.format(param_name))

        self._param_dict[param_name] = (param_name,
                                        param_latex,
                                        fget.__get__(self),
                                        fset.__get__(self),
                                        default_mode,
                                        default_fit,
                                        default_bounds)

    def compile_fitparams(self):
        """
        Loops through and finds all fitting parameters in the class and adds
        it to the internal dictionary
        """
        for fitparams in self.find_fitparams():

            get_func = fitparams.fget
            set_func = fitparams.fset
            param_name = get_func.param_name
            param_latex = get_func.param_latex
            def_mode = get_func.default_mode
            def_fit = get_func.default_fit
            def_bounds = get_func.default_bounds
            self.add_fittable_param(param_name, param_latex, get_func,
                                    set_func, def_mode, def_fit, def_bounds)

    def modify_bounds(self, parameter, new_bounds):
        """
        Modifies the fitting boundary of a parameter

        Parameters
        ----------
        parameter : str
            Name of parameter (given by ``param_name`` in :func:`fitparam`)

        new_bounds : :obj:`list`
            New minimum and maximum fitting boundaries.

        """
        name, latex, fget, fset, mode, to_fit, bounds = \
            self._param_dict[parameter]

        bounds = new_bounds

        self._param_dict[parameter] = \
            name, latex, fget, fset, mode, to_fit, bounds

    def __getitem__(self, key):
        param = self._param_dict[key]

        return param[2]()

    def __setitem__(self, key, value):
        return self._param_dict[key][3](value)

    def find_fitparams(self):
        """
        Finds and returns fitting parameters

        Yields
        ------
        method : function
            class method that is defined with the :func:`fitparam` decorator
        """

        for klass in self.__class__.mro():

            for method in klass.__dict__.values():
                if hasattr(method, 'fget'):
                    prop = method.fget
                    if hasattr(prop, 'decorated'):
                        if prop.decorated == 'fitparam':
                            yield method

    def fitting_parameters(self):
        """
        Returns all fitting parameters found as a dictionary

        Returns
        -------
        params : :obj:`dict`
            Dictionary with key as the parameter name (``param_name``) 
            and value as a tuple with:
                * parameter name
                * parameter name in Latex form
                * get function
                * set function
                * fitting scale
                * fit as default
                * fitting boundaries

        """
        return self._param_dict



from .optimizer import Optimizer
import pymultinest
import numpy as np
import os
from taurex.mpi import get_rank, barrier
from taurex.util.util import read_table, read_error_line, read_error_into_dict, quantile_corner, recursively_save_dict_contents_to_output
import random
from taurex.util.util import weighted_avg_and_std
from taurex import OutputSize


class MultiNestOptimizer(Optimizer):

    def __init__(self, multi_nest_path=None, observed=None, model=None,
                 sampling_efficiency='parameter',
                 num_live_points=1500,
                 max_iterations=0,
                 search_multi_modes=True,
                 num_params_cluster=None,
                 maximum_modes=100,
                 constant_efficiency_mode=False,
                 evidence_tolerance=0.5,
                 mode_tolerance=-1e90,
                 importance_sampling=False,
                 resume=False,
                 multinest_prefix='1-',
                 verbose_output=True, sigma_fraction=0.1):
        super().__init__('Multinest', observed, model, sigma_fraction)

        # sampling chains directory
        self.nest_path = 'chains/'
        self.nclust_par = -1
        # sampling efficiency (parameter, ...)
        self.sampling_eff = sampling_efficiency
        # number of live points
        self.n_live_points = int(num_live_points)
        # maximum no. of iterations (0=inf)
        self.max_iter = int(max_iterations)
        # search for multiple modes
        self.multimodes = search_multi_modes
        # parameters on which to cluster, e.g. if nclust_par = 3, it will
        # cluster on the first 3 parameters only.
        # if ncluster_par = -1 it clusters on all parameters
        self.nclust_par = num_params_cluster
        # maximum number of modes
        self.max_modes = int(maximum_modes)
        # run in constant efficiency mode
        self.const_eff = constant_efficiency_mode
        # set log likelihood tolerance. If change is smaller, multinest will have converged
        self.evidence_tolerance = evidence_tolerance
        self.mode_tolerance = mode_tolerance
        # importance nested sampling
        self.imp_sampling = importance_sampling
        if self.imp_sampling:
            self.multimodes = False

        self.multinest_prefix = multinest_prefix

        self.dir_multinest = multi_nest_path
        if get_rank() == 0:

            if not os.path.exists(self.dir_multinest):
                self.info('Directory %s does not exist, creating',
                          self.dir_multinest)
                os.makedirs(self.dir_multinest)
        barrier()

        self.info('Found directory %s', self.dir_multinest)

        self.resume = resume
        self.verbose = verbose_output

    def compute_fit(self):

        data = self._observed.spectrum
        datastd = self._observed.errorBar
        sqrtpi = np.sqrt(2*np.pi)

        def multinest_loglike(cube, ndim, nparams):
            # log-likelihood function called by multinest
            fit_params_container = np.array(
                [cube[i] for i in range(len(self.fitting_parameters))])
            chi_t = self.chisq_trans(fit_params_container, data, datastd)

            # print('---------START---------')
            # print('chi_t',chi_t)
            # print('LOG',loglike)
            loglike = -np.sum(np.log(datastd*sqrtpi)) - 0.5 * chi_t
            # print(loglike)
            return loglike

        def multinest_uniform_prior(cube, ndim, nparams):
            # prior distributions called by multinest. Implements a uniform prior
            # converting parameters from normalised grid to uniform prior
            # print(type(cube))
            for idx, bounds in enumerate(self.fit_boundaries):
                # print(idx,self.fitting_parameters[idx])
                bound_min, bound_max = bounds
                cube[idx] = (cube[idx] * (bound_max-bound_min)) + bound_min
                #print('CUBE idx',cube[idx])
            # print('-----------')
        # status = None
        # def dump_call(nSamples,nlive,nPar,physLive,posterior,paramConstr,maxloglike,logZ,INSlogZ,logZerr,context):
        #    status = (nSamples,nlive,nPar,physLive,posterior,paramConstr,maxloglike,logZ,INSlogZ,logZerr,context)

        ndim = len(self.fitting_parameters)
        self.warning('Number of dimensions {}'.format(ndim))
        self.warning('Fitting parameters {}'.format(self.fitting_parameters))

        ncluster = self.nclust_par

        if isinstance(ncluster, float):
            ncluster = int(ncluster)

        if ncluster is not None and ncluster <= 0:
            ncluster = None

        if ncluster is None:
            self.nclust_par = ndim  # For writing to output later on

        self.info('Beginning fit......')
        pymultinest.run(LogLikelihood=multinest_loglike,
                        Prior=multinest_uniform_prior,
                        n_dims=ndim,
                        multimodal=self.multimodes,
                        n_clustering_params=ncluster,
                        max_modes=self.max_modes,
                        outputfiles_basename=os.path.join(self.dir_multinest,
                                                          self.multinest_prefix),
                        const_efficiency_mode=self.const_eff,
                        importance_nested_sampling=self.imp_sampling,
                        resume=self.resume,
                        verbose=self.verbose,
                        sampling_efficiency=self.sampling_eff,
                        evidence_tolerance=self.evidence_tolerance,
                        mode_tolerance=self.mode_tolerance,
                        n_live_points=self.n_live_points,
                        max_iter=self.max_iter
                        )

        self.info('Fit complete.....')

        self._multinest_output = self.store_nest_solutions()
        self.debug('Multinest output %s', self._multinest_output)

    def write_optimizer(self, output):
        opt = super().write_optimizer(output)

        # sampling efficiency (parameter, ...)
        opt.write_scalar('sampling_eff ', self.sampling_eff)
        # number of live points
        opt.write_scalar('num_live_points', self.n_live_points)
        # maximum no. of iterations (0=inf)
        opt.write_scalar('max_iterations', self.max_iter)
        # search for multiple modes
        opt.write_scalar('search_multimodes', self.multimodes)
        # parameters on which to cluster, e.g. if nclust_par = 3, it will cluster on the first 3 parameters only.
        # if ncluster_par = -1 it clusters on all parameters
        opt.write_scalar('nclust_parameter', self.nclust_par)
        # maximum number of modes
        opt.write_scalar('max_modes', self.max_modes)
        # run in constant efficiency mode
        opt.write_scalar('constant_efficiency', self.const_eff)
        # set log likelihood tolerance. If change is smaller, multinest will have converged
        opt.write_scalar('evidence_tolerance', self.evidence_tolerance)
        opt.write_scalar('mode_tolerance', self.mode_tolerance)
        # importance nested sampling
        opt.write_scalar('importance_sampling ', self.imp_sampling)

        return opt

    def write_fit(self, output):
        fit = super().write_fit(output)

        if self._multinest_output:
            recursively_save_dict_contents_to_output(
                output, self._multinest_output)

        return fit

    # Laziness brought us to this point
    # This function is so big and I cannot be arsed to rewrite this in a nicer way, if some angel does it
    # for me then I will buy them TWO beers.
    def store_nest_solutions(self):

        self.warning('Store the multinest results')
        NEST_out = {'solutions': {}}
        data = np.loadtxt(os.path.join(self.dir_multinest,
                                       '{}.txt'.format(self.multinest_prefix)))

        NEST_analyzer = pymultinest.Analyzer(n_params=len(
            self.fitting_parameters), outputfiles_basename=os.path.join(self.dir_multinest, self.multinest_prefix))
        NEST_stats = NEST_analyzer.get_stats()
        NEST_out['NEST_stats'] = NEST_stats
        NEST_out['global_logE'] = (
            NEST_out['NEST_stats']['global evidence'], NEST_out['NEST_stats']['global evidence error'])

        # Bypass if run in multimodes = False. Pymultinest.Analyser does not report means and sigmas in this case
        if len(NEST_out['NEST_stats']['modes']) == 0:
            with open(os.path.join(self.dir_multinest, '{}stats.dat'.format(self.multinest_prefix))) as file:
                lines = file.readlines()
            stats = {'modes': []}
            read_error_into_dict(lines[0], stats)

            text = ''.join(lines[2:])
            # without INS:
            parts = text.split("\n\n")
            mode = {'index': 0}

            modelines = parts[0]
            t = read_table(modelines, title='Parameter')
            mode['mean'] = t[:, 1].tolist()
            mode['sigma'] = t[:, 2].tolist()

            mode['maximum'] = read_table(parts[1], title='Parameter', d=None)[
                :, 1].tolist()
            mode['maximum a posterior'] = read_table(
                parts[2], title='Parameter', d=None)[:, 1].tolist()

            if 'Nested Sampling Global Log-Evidence'.lower() in stats:
                mode['Local Log-Evidence'.lower()
                     ] = stats['Nested Sampling Global Log-Evidence'.lower()]
                mode['Local Log-Evidence error'.lower()
                     ] = stats['Nested Sampling Global Log-Evidence error'.lower()]
            else:
                mode['Local Log-Evidence'.lower()
                     ] = stats['Nested Importance Sampling Global Log-Evidence'.lower()]
                mode['Local Log-Evidence error'.lower()
                     ] = stats['Nested Importance Sampling Global Log-Evidence error'.lower()]

            mode['Strictly Local Log-Evidence'.lower()
                 ] = mode['Local Log-Evidence'.lower()]
            mode['Strictly Local Log-Evidence error'.lower()
                 ] = mode['Local Log-Evidence error'.lower()]

            NEST_out['NEST_stats']['modes'] = [mode]

        modes = []
        modes_weights = []
        chains = []
        chains_weights = []

        if self.multimodes:

            # separate modes. get individual samples for each mode

            # get parameter values and sample probability (=weight) for each mode
            with open(os.path.join(self.dir_multinest, '{}post_separate.dat'.format(self.multinest_prefix))) as f:
                lines = f.readlines()
                for idx, line in enumerate(lines):
                    if idx > 2:  # skip the first two lines
                        if lines[idx-1] == '\n' and lines[idx-2] == '\n':
                            modes.append(chains)
                            modes_weights.append(chains_weights)
                            chains = []
                            chains_weights = []
                    chain = [float(x) for x in line.split()[2:]]
                    if len(chain) > 0:
                        chains.append(chain)
                        chains_weights.append(float(line.split()[0]))
                modes.append(chains)
                modes_weights.append(chains_weights)
            modes_array = []
            for mode in modes:
                mode_array = np.zeros((len(mode), len(mode[0])))
                for idx, line in enumerate(mode):
                    mode_array[idx, :] = line
                modes_array.append(mode_array)
        else:
            # not running in multimode. Get chains directly from file 1-.txt
            modes_array = [data[:, 2:]]
            chains_weights = [data[:, 0]]
            modes_weights.append(chains_weights[0])
            modes = [0]

        modes_weights = np.asarray(modes_weights)
        for nmode in range(len(modes)):
            self.debug('Nmode: {}'.format(nmode))

            mydict = {'type': 'nest',
                      'local_logE': (NEST_out['NEST_stats']['modes'][0]['local log-evidence'],  NEST_out['NEST_stats']['modes'][0]['local log-evidence error']),
                      'weights': np.asarray(modes_weights[nmode]),
                      'tracedata': modes_array[nmode],
                      'fit_params': {}}

            for idx, param_name in enumerate(self.fit_names):

                trace = modes_array[nmode][:, idx]
                q_16, q_50, q_84 = quantile_corner(trace, [0.16, 0.5, 0.84],
                                                   weights=np.asarray(modes_weights[nmode]))
                mydict['fit_params'][param_name] = {
                    'value': q_50,
                    'sigma_m': q_50-q_16,
                    'sigma_p': q_84-q_50,
                    'nest_map': NEST_stats['modes'][nmode]['maximum a posterior'][idx],
                    'mean': NEST_stats['modes'][nmode]['mean'][idx],
                    'nest_sigma': NEST_stats['modes'][nmode]['sigma'][idx],
                    'trace': trace,
                }

            NEST_out['solutions']['solution{}'.format(nmode)] = mydict

        return NEST_out

    def generate_solution(self, output_size=OutputSize.heavy):

        solution = super().generate_solution(output_size=output_size)

        solution['GlobalStats'] = self._multinest_output['NEST_stats']
        return solution

    def sample_parameters(self, solution):
        from taurex.util.util import random_int_iter
        solution_id = 'solution{}'.format(solution)
        samples = self._multinest_output['solutions'][solution_id]['tracedata']
        weights = self._multinest_output['solutions'][solution_id]['weights']

        for x in random_int_iter(samples.shape[0], self._sigma_fraction):
            w = weights[x]+1e-300

            yield samples[x, :], w

    def get_solution(self):
        names = self.fit_names
        opt_values = self.fit_values
        opt_map = self.fit_values
        solutions = [
            (k, v) for k, v in self._multinest_output['solutions'].items() if 'solution' in k]
        for k, v in solutions:
            solution_idx = int(k[8:])
            for p_name, p_value in v['fit_params'].items():
                if p_name in ('mu_derived',):
                    continue
                idx = names.index(p_name)
                opt_map[idx] = p_value['nest_map']
                opt_values[idx] = p_value['value']

            yield solution_idx, opt_map, opt_values, [
                ('Statistics', {'local log-evidence': self._multinest_output['NEST_stats']['modes'][solution_idx]['local log-evidence'],
                                'local log-evidence error': self._multinest_output['NEST_stats']['modes'][solution_idx]['local log-evidence error']}),
                ('fit_params', v['fit_params']),
                ('tracedata', v['tracedata']),
                ('weights', v['weights'])]



"""General utility functions"""

from taurex.constants import AMU
import math
import re
import numpy as np
mass = {
  "H":	1.00794,
  "He":	4.002602,
  "Li":	6.941,
  "Be":	9.012182,
  "B":	10.811,
  "C":	12.011,
  "N":	14.00674,
  "O":	15.9994,
  "F":	18.9984032,
  "Ne":	20.1797,
  "Na":	22.989768,
  "Mg":	24.3050,
  "Al":	26.981539,
  "Si":	28.0855,
  "P":	30.973762,
  "S":	32.066,
  "Cl":	35.4527,
  "Ar":	39.948,
  "K":	39.0983,
  "Ca":	40.078,
  "Sc":	44.955910,
  "Ti":	47.88,
  "V":	50.9415,
  "Cr":	51.9961,
  "Mn":	54.93805,
  "Fe":	55.847,
  "Co":	58.93320,
  "Ni":	58.6934,
  "Cu":	63.546,
  "Zn":	65.39,
  "Ga":	69.723,
  "Ge":	72.61,
  "As":	74.92159,
  "Se":	78.96,
  "Br":	79.904,
  "Kr":	83.80,
  "Rb":	85.4678,
  "Sr":	87.62,
  "Y":	88.90585,
  "Zr":	91.224,
  "Nb":	92.90638,
  "Mo":	95.94,
  "Tc":	98,
  "Ru":	101.07,
  "Rh":	102.90550,
  "Pd":	106.42,
  "Ag":	107.8682,
  "Cd":	112.411,
  "In":	114.82,
  "Sn":	118.710,
  "Sb":	121.757,
  "Te":	127.60,
  "I":	126.90447,
  "Xe":	131.29,
  "Cs":	132.90543,
  "Ba":	137.327,
  "La":	138.9055,
  "Ce":	140.115,
  "Pr":	140.90765,
  "Nd":	144.24,
  "Pm":	145,
  "Sm":	150.36,
  "Eu":	151.965,
  "Gd":	157.25,
  "Tb":	158.92534,
  "Dy":	162.50,
  "Ho":	164.93032,
  "Er":	167.26,
  "Tm":	168.93421,
  "Yb":	173.04,
  "Lu":	174.967,
  "Hf":	178.49,
  "Ta":	180.9479,
  "W":	183.85,
  "Re":	186.207,
  "Os":	190.2,
  "Ir":	192.22,
  "Pt":	195.08,
  "Au":	196.96654,
  "Hg":	200.59,
  "Tl":	204.3833,
  "Pb":	207.2,
  "Bi":	208.98037,
  "Po":	209,
  "At":	210,
  "Rn":	222,
  "Fr":	223,
  "Ra":	226.0254,
  "Ac":	227,
  "Th":	232.0381,
  "Pa":	213.0359,
  "U":	238.0289,
  "Np":	237.0482,
  "Pu":	244,
  "Am":	243,
  "Cm":	247,
  "Bk":	247,
  "Cf":	251,
  "Es":	252,
  "Fm":	257,
  "Md":	258,
  "No":	259,
  "Lr":	260,
  "Rf":	261,
  "Db":	262,
  "Sg":	263,
  "Bh":	262,
  "Hs":	265,
  "Mt":	266,
}

def calculate_weight(chem):
    s = re.findall('([A-Z][a-z]?)([0-9]*)', chem)
    compoundweight = 0

    for element, count in s:
        count = int(count or '1')
        compoundweight += mass[element] * count
    return compoundweight


_mol_latex={

    'HE':
        'He',
    'H2':
        'H$_2$',
    'N2':
        'N$_2$',
    'O2':
        'O$_2$',
    'CO2':
        'CO$_2$',
    'CH4':
        'CH$_4$',
    'CO':
        'CO',
    'NH3':
        'NH$_3$',
    'H2O':
        'H$_2$O',
    'C2H2':
        'C$_2$H$_2$',
    'HCN':
        'HCN',
    'H2S':
        'H$_2$S',
    'SIO2':
        'SiO$_2$',
    'SO2':
        'SO$_2$',
}
"""Latex versions of molecule names"""

def get_molecular_weight(gasname):
    """
    For a given molecule return the molecular weight in atomic units
    
    Parameters
    ----------
    gasname : str
        Name of molecule
    
    Returns
    -------
    float : 
        molecular weight in amu or 0 if not found
    
    """
    mu = calculate_weight(gasname)


    return mu * AMU


def molecule_texlabel(gasname):
    """
    For a given molecule return its latex form
    
    Parameters
    ----------
    gasname : str
        Name of molecule
    
    Returns
    -------
    str : 
        Latex form of the molecule or just the passed name if not found

    
    """
    gasname = gasname

    try:
        return _mol_latex[gasname]
    except KeyError:
        return gasname


def bindown(original_bin,original_data,new_bin,last_point=None):
    """
    This method quickly bins down by taking the mean.
    The numpy histogram function is exploited to do this quickly
    
    Parameters
    ----------
    original_bin: :obj:`numpy.array`
        The original bins for the that we want to bin down
    
    original_data: :obj:`numpy.array`
        The associated data that will be averaged along the new bins

    new_bin: :obj:`numpy.array`
        The new binnings we want to use (must have less points than the original)
    
    Returns
    -------
    :obj:`array`
        Binned mean of ``original_data``

    
    """
    import numpy as np
    #print(original_bin.shape,original_data.shape)
    #if last_point is None:
    #    last_point = new_bin[-1]*2
    
    #calc_bin = np.append(new_bin,last_point)
    #return(np.histogram(original_bin, calc_bin, weights=original_data)[0] /
    #          np.histogram(original_bin,calc_bin)[0])


    
    filter_lhs = np.zeros(new_bin.shape[0]+1)
    filter_lhs[0] = new_bin[0]
    filter_lhs[0] -= (new_bin[1] - new_bin[0])/2
    filter_lhs[-1] = new_bin[-1]
    filter_lhs[-1] += (new_bin[-1] - new_bin[-2])/2
    filter_lhs[1:-1] = (new_bin[1:] + new_bin[:-1])/2
    axis = len(original_data.shape)-1
    if axis:
        digitized = np.digitize(original_bin,filter_lhs,right=True)
        axis = len(original_data.shape)-1
        bin_means = [original_data[...,digitized == i].mean(axis=axis) for i in range(1, len(filter_lhs))]
        return np.column_stack(bin_means)
    return np.histogram(original_bin, filter_lhs, weights=original_data)[0]/np.histogram(original_bin,filter_lhs)[0]

def movingaverage(a, n=3) :
    """
    Computes moving average

    Parameters
    ----------
    a : :obj:`array`
        Array to compute average
    
    n : int
        Averaging window

    Returns
    -------
    :obj:`array`
        Resultant array

    """
    import numpy as np
    ret = np.cumsum(a)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

def quantile_corner(x, q, weights=None):
    """

    * Taken from corner.py
    __author__ = "Dan Foreman-Mackey (danfm@nyu.edu)"
    __copyright__ = "Copyright 2013-2015 Daniel Foreman-Mackey"

    Like numpy.percentile, but:
    
    * Values of q are quantiles [0., 1.] rather than percentiles [0., 100.]
    * scalar q not supported (q must be iterable)
    * optional weights on x

    Parameters
    ----------

    x : :obj:`array`
        Input array or object that can be converted to an array.
    
    q : :obj:`array` or float
        Percentile or sequence of percentiles to compute, which must be between 0 and 1 inclusive.
    
    weights : :obj:`array` or float , optional
        Weights on x

    Returns
    -------
    percentile : scalar or ndarray
        

    """
    import numpy as np
    if weights is None:
        return np.percentile(x, [100. * qi for qi in q])
    else:
        idx = np.argsort(x)
        xsorted = x[idx]
        cdf = np.add.accumulate(weights[idx])
        cdf /= cdf[-1]
        return np.interp(q, cdf, xsorted).tolist()

def loadtxt2d(intext):
    """
    
    Wraps loadtext and either returns a 2d array
    or 1d array

    """

    try:
        return np.loadtxt(intext, ndmin=2)
    except:
        return np.loadtxt(intext)

def read_error_line(l):
    """ 
    Reads line?
    """
    print('_read_error_line -> line>', l)
    name, values = l.split('   ', 1)
    print('_read_error_line -> name>', name)
    print('_read_error_line -> values>', values)
    name = name.strip(': ').strip()
    values = values.strip(': ').strip()
    v, error = values.split(" +/- ")
    return name, float(v), float(error)

def read_error_into_dict(l, d):
    """
    Reads the error into dict?
    """
    name, v, error = read_error_line(l)
    d[name.lower()] = v
    d['%s error' % name.lower()] = error

def read_table(txt, d=None, title=None):
    """
    Yeah whatever i give up
    """
    from io import StringIO
    import numpy as np
    if title is None:
        title, table = txt.split("\n", 1)
    else:
        table = txt
    header, table = table.split("\n", 1)
    data = loadtxt2d(StringIO(table))
    if d is not None:
        d[title.strip().lower()] = data
    if len(data.shape) == 1:
        data = np.reshape(data, (1, -1))
    return data

def decode_string_array(f):
    """Helper to decode strings from hdf5"""
    sl = list(f)
    return [s[0].decode('utf-8') for s in sl] 


def recursively_save_dict_contents_to_output(output, dic):
    """
    Will recursive write a dictionary into output.

    Parameters
    ----------
    output : :class:`~taurex.output.output.Output` or :class:`~taurex.output.output.OutputGroup`
        Group (or root) in output file to write to
    
    dic : :obj:`dict`
        Dictionary we want to write

    """
    import numpy as np

    for key, item in dic.items():
        
        

            try:
                store_thing(output, key, item)
            except TypeError:
                raise ValueError('Cannot save %s type'%type(item))


def store_thing(output, key, item):
    if isinstance(item, (float,int,np.int64,np.float64,)):
        output.write_scalar(key,item)
    elif isinstance(item,(np.ndarray,)):
        output.write_array(key,item)
    elif isinstance(item,(str,)):
        output.write_string(key,item)
    elif isinstance(item,(list,tuple,)):
        if isinstance(item,tuple):
            item = list(item)
        if True in [isinstance(x,str) for x in item]:
            output.write_string_array(key,item)
        else:
            try:
                output.write_array(key,np.array(item))
                
            except TypeError:
                for idx,val in enumerate(item):
                    new_key = '{}{}'.format(key,idx)
                    store_thing(output,new_key,val)

    elif isinstance(item, dict):
            group = output.create_group(key)
            recursively_save_dict_contents_to_output(group, item)
    else:
        raise TypeError


def weighted_avg_and_std(values, weights, axis=None):
    """
    Computes weight average and standard deviation

    Parameters
    ----------
    values : :obj:`array`
        Input array
    
    weights : :obj:`array`
        Must be same shape as ``values``
        
    
    axis : int , optional
        axis to perform weighting
    
    """


    import numpy as np
    average = np.average(values, weights=weights,axis=axis)
    variance = np.average((values-average)**2, weights=weights, axis=axis)  # Fast and numerically precise
    return (average, np.sqrt(variance))


def random_int_iter(total, fraction):
    import random
    n_points = int(total*fraction)

    samples = random.sample(range(total), n_points)
    for x in samples:
        yield x


def compute_bin_edges(wngrid):
    import numpy as np
    diff = np.diff(wngrid)/2
    edges = np.concatenate([[wngrid[0]-(wngrid[1]-wngrid[0])/2], 
                           wngrid[:-1]+diff, [(wngrid[-1]-wngrid[-2])/2 +
                           wngrid[-1]]])
    return edges, np.abs(np.diff(edges))


def clip_native_to_wngrid(native_grid, wngrid):

    min_wngrid = wngrid.min()
    max_wngrid = wngrid.max()
    #Compute the maximum width
    wnwidths = compute_bin_edges(wngrid)[-1]
    wn_min = min_wngrid - wnwidths.max()
    wn_max = max_wngrid + wnwidths.max()

    native_filter = (native_grid >= wn_min) & (native_grid <= wn_max)
    return native_grid[native_filter]

def wnwidth_to_wlwidth(wngrid, wnwidth):
    return 10000*wnwidth/(wngrid**2)



def class_for_name(module_name, class_name):
    import importlib
    # load the module, will raise ImportError if module cannot be loaded
    m = importlib.import_module(module_name)
    # get the class, will raise AttributeError if class cannot be found
    c = getattr(m, class_name)
    return c

def create_grid_res(resolution, wave_min, wave_max):
    #
    # R = l/Dl
    # l = (l-1)+Dl/2 + (Dl-1)/2
    #
    # --> (R - 1/2)*Dl = (l-1) + (Dl-1)/2
    #
    # 
    wave_list = []
    width_list = []
    wave = wave_min
    width = wave/resolution    
    
    while wave < wave_max:
        width = wave / (resolution - 0.5) + width/2/(resolution - 0.5)
        wave = resolution * width 
        width_list.append(width)
        wave_list.append(wave)

    return np.array((wave_list ,width_list)).T


from taurex.log import Logger
from taurex.util import get_molecular_weight
from taurex.data.fittable import Fittable
import numpy as np
from taurex.output.writeable import Writeable
from taurex.cache import OpacityCache


class Chemistry(Fittable, Logger, Writeable):
    """
    *Abstract Class*

    Skeleton for defining chemistry. Must implement
    methods:

    - :func:`activeGases`
    - :func:`inactiveGases`
    - :func:`activeGasMixProfile`
    - :func:`inactiveGasMixProfile`

    *Active* are those that are actively
    absorbing in the atmosphere. In technical terms they are molecules
    that have absorption cross-sections. You can see which molecules
    are able to actively absorb by doing:
    You can find out what molecules can actively absorb by doing:

            >>> avail_active_mols = OpacityCache().find_list_of_molecules()

    Parameters
    ----------
    name : str
        Name used in logging

    """

    def __init__(self, name):
        Logger.__init__(self, name)
        Fittable.__init__(self)

        self.mu_profile = None
        self._avail_active = OpacityCache().find_list_of_molecules()

    @property
    def availableActive(self):
        """
        Returns a list of available
        actively absorbing molecules

        Returns
        -------
        molecules: :obj:`list`
            Actively absorbing molecules
        """
        return self._avail_active

    @property
    def activeGases(self):
        """
        **Requires implementation**

        Should return a list of molecule names

        Returns
        -------
        active : :obj:`list`
            List of active gases

        """
        raise NotImplementedError

    @property
    def inactiveGases(self):
        """
        **Requires implementation**

        Should return a list of molecule names

        Returns
        -------
        inactive : :obj:`list`
            List of inactive gases

        """
        raise NotImplementedError

    def initialize_chemistry(self, nlayers=100, temperature_profile=None,
                             pressure_profile=None, altitude_profile=None):
        """
        **Requires implementation**

        Derived classes should implement this to compute the active and
        inactive gas profiles

        Parameters
        ----------
        nlayers: int
            Number of layers in atmosphere

        temperature_profile: :obj:`array`
            Temperature profile in K, must have length ``nlayers``

        pressure_profile: :obj:`array`
            Pressure profile in Pa, must have length ``nlayers``

        altitude_profile: :obj:`array`
            Altitude profile in m, must have length ``nlayers``

        """
        self.compute_mu_profile(nlayers)

    @property
    def activeGasMixProfile(self):
        """
        **Requires implementation**

        Should return profiles of shape ``(nactivegases,nlayers)``. Active
        refers to gases that are actively absorbing in the atmosphere.
        Another way to put it these are gases where molecular cross-sections
        are used.

        """

        raise NotImplementedError

    @property
    def inactiveGasMixProfile(self):
        """
        **Requires implementation**

        Should return profiles of shape ``(ninactivegases,nlayers)``.
        These general refer to gases: ``H2``, ``He`` and ``N2``


        """
        raise NotImplementedError

    @property
    def muProfile(self):
        """
        Molecular weight for each layer of atmosphere


        Returns
        -------
        mix_profile : :obj:`array`

        """
        return self.mu_profile

    def get_gas_mix_profile(self, gas_name):
        """
        Returns the mix profile of a particular gas

        Parameters
        ----------
        gas_name : str
            Name of gas

        Returns
        -------
        mixprofile : :obj:`array`
            Mix profile of gas with shape ``(nlayer)``

        """
        if gas_name in self.activeGases:
            idx = self.activeGases.index(gas_name)
            return self.activeGasMixProfile[idx]
        elif gas_name in self.inactiveGases:
            idx = self.inactiveGases.index(gas_name)
            return self.inactiveGasMixProfile[idx]
        else:
            raise KeyError

    def compute_mu_profile(self, nlayers):
        """
        Computes molecular weight of atmosphere
        for each layer

        Parameters
        ----------
        nlayers: int
            Number of layers
        """

        self.mu_profile = np.zeros(shape=(nlayers,))
        if self.activeGasMixProfile is not None:
            for idx, gasname in enumerate(self.activeGases):
                self.mu_profile += self.activeGasMixProfile[idx] * \
                    get_molecular_weight(gasname)
        if self.inactiveGasMixProfile is not None:
            for idx, gasname in enumerate(self.inactiveGases):
                self.mu_profile += self.inactiveGasMixProfile[idx] * \
                    get_molecular_weight(gasname)

    def write(self, output):
        """
        Writes chemistry class and arguments to file

        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`

        """
        gas_entry = output.create_group('Chemistry')
        gas_entry.write_string('chemistry_type', self.__class__.__name__)
        gas_entry.write_string_array('active_gases', self.activeGases)
        gas_entry.write_string_array('inactive_gases', self.inactiveGases)
        return gas_entry




from .star import BlackbodyStar
import numpy as np
import os
from taurex.constants import MSOL
import math


class PhoenixStar(BlackbodyStar):
    """
    A star that uses the `PHOENIX <https://www.aanda.org/articles/aa/abs/2013/05/aa19058-12/aa19058-12.html>`_
    synthetic stellar atmosphere spectrums.

    These spectrums are read from ``.gits.gz`` files in a directory given by
    ``phoenix_path``
    Each file must contain the spectrum for one temperature

    Parameters
    ----------

    phoenix_path: str, **required**
        Path to folder containing phoenix ``fits.gz`` files

    temperature: float, optional
        Stellar temperature in Kelvin

    radius: float, optional
        Stellar radius in Solar radius

    metallicity: float, optional
        Metallicity in solar values

    mass: float, optional
        Stellar mass in solar mass

    distance: float, optional
        Distance from Earth in pc

    magnitudeK: float, optional
        Maginitude in K band


    Raises
    ------
    Exception
        Raised when no phoenix path is defined


    """

    def __init__(self, temperature=5000, radius=1.0, metallicity=1.0, mass=1.0,
                 distance=1, magnitudeK=10.0, phoenix_path=None):
        super().__init__(temperature=temperature, radius=radius,
                         distance=distance,
                         magnitudeK=magnitudeK, mass=mass,
                         metallicity=metallicity)
        if phoenix_path is None:
            self.error('No file path to phoenix files defined')
            raise Exception('No file path to phoenix files defined')

        self.info('Star is PHOENIX type')
        self._phoenix_path = phoenix_path

        self.get_avail_phoenix()
        self.use_blackbody = False
        self.recompute_spectra()
        # self.preload_phoenix_spectra()

    def compute_logg(self):
        """
        Computes log(surface_G)

        """
        import astropy.units as u
        from astropy.constants import G
        mass = self._mass * u.kg
        radius = self._radius * u.m

        small_g = (G * mass) / (radius**2)

        small_g = small_g.to(u.cm/u.s**2)

        return math.log10(small_g.value)

    def recompute_spectra(self):

        if self.temperature > self._T_list.max() or \
                self.temperature < self._T_list.min():

            self.use_blackbody = True
        else:
            self.use_blackbody = False
            self._logg = self.compute_logg()
            f = self.find_nearest_file()
            self.read_spectra(f)

    def read_spectra(self, p_file):
        from astropy.io import fits
        import astropy.units as u

        with fits.open(p_file) as hdu:
            strUnit = hdu[1].header['TUNIT1']
            wl = hdu[1].data.field('Wavelength') * u.Unit(strUnit)

            strUnit = hdu[1].header['TUNIT2']
            sed = hdu[1].data.field('Flux') * u.Unit(strUnit)

            self.wngrid = 10000/(wl.value)
            argidx = np.argsort(self.wngrid)
            self._base_sed = sed.to(u.W/u.m**2/u.micron)
            self.wngrid = self.wngrid[argidx]
            self._base_sed = self._base_sed[argidx]

    @property
    def temperature(self):
        """
        Effective Temperature in Kelvin

        Returns
        -------
        T: float

        """
        return self._temperature

    @temperature.setter
    def temperature(self, value):
        self._temperature = value
        self.recompute_spectra()

    @property
    def mass(self):
        """
        Mass of star in solar mass

        Returns
        -------
        M: float

        """
        return self._mass

    @mass.setter
    def mass(self, value):
        self._mass = value * MSOL
        self.recompute_spectra()

    def find_nearest_file(self):

        idx = self._index_finder(
            [self._temperature, self._logg, self._metallicity])
        return self._files[int(idx)]

    def get_avail_phoenix(self):
        from scipy.interpolate import NearestNDInterpolator
        import glob
        files = glob.glob(os.path.join(self._phoenix_path, '*.spec.fits.gz'))
        self._files = files
        self._T_list = np.array(
            [np.float(os.path.basename(k)[3:8]) for k in files])*100
        self._Logg_list = np.array(
            [np.float(os.path.basename(k)[9:12]) for k in files])
        self._Z_list = np.array(
            [np.float(os.path.basename(k)[13:16]) for k in files])
        self._index_finder = NearestNDInterpolator(
            (self._T_list, self._Logg_list, self._Z_list),
            np.arange(0, self._T_list.shape[0]))

    # def preload_phoenix_spectra(self):

    #     T_list = self.detect_all_T(self._phoenix_path)

    #     self._temperature_grid = np.array([x[0] for x in T_list])
    #     self.debug('Detected temepratures = %s',self._temperature_grid)
    #     self._avail_max_temp = max(self._temperature_grid)
    #     self._avail_min_temp = min(self._temperature_grid)

    #     self.debug('Temperature range = [%s-%s] ',self._avail_min_temp,self._avail_max_temp)
    #     self._max_index = self._temperature_grid.shape[0]
    #     self.spectra_grid = []

    #     #Load in all arrays
    #     for temp,f in T_list:
    #         self.debug('Loading %s %s',temp,f)
    #         arr = np.loadtxt(f)
    #         grid = 10000/np.copy(arr[:,0])
    #         sorted_idx = np.argsort(grid)

    #         self.wngrid = 10000/np.copy(arr[sorted_idx,0])
    #         self.spectra_grid.append(arr[sorted_idx,1]*10.0) #To SI

    # def find_closest_index(self,T):
    #     """
    #     Finds the two closest indices in our temeprature grid to our desired temperature

    #     Parameters
    #     ----------
    #     T: float
    #         Temperature in Kelvin

    #     Returns
    #     -------
    #     t_min: int
    #         Index to the left of ``T``

    #     t_max: int
    #         Index to the right of ``T``

    #     """

    #     t_min=self._temperature_grid.searchsorted(T,side='right')-1
    #     t_max = t_min+1

    #     return t_min,t_max

    # def interpolate_linear_temp(self,T):
    #     """
    #     Linearly interpolates the spectral emission density grid to the
    #     temperature given by ``T``

    #     Parameters
    #     ----------
    #     T: float
    #         Temeprature to interpolate to

    #     Returns
    #     -------
    #     out: :obj:`array`
    #         Spectral emission density interpolated to desired temperature

    #     """
    #     t_min,t_max = self.find_closest_index(T)
    #     if self._temperature_grid[t_min] == T:
    #         return self.spectra_grid[t_min]

    #     Tmax = self._temperature_grid[t_max]
    #     Tmin = self._temperature_grid[t_min]
    #     fx0=self.spectra_grid[t_min]
    #     fx1 = self.spectra_grid[t_max]

    #     return interp_lin_only(fx0,fx1,T,Tmin,Tmax)

    # def detect_all_T(self,path):
    #     """
    #     Finds files and detects all temperatures available in path

    #     Parameters
    #     ----------
    #     path: str
    #         Path to directory containing PHOENIX data

    #     """
    #     files = glob.glob(os.path.join(self._phoenix_path,'*.fits.gz'))
    #     files.sort()

    #     temp_list = []
    #     for f in files:
    #         #Gewt just the name of the file
    #         clean_name = pathlib.Path(f).stem
    #         #Split it by numbers
    #         split = re.split('(\d+)',clean_name)
    #         try:
    #             _T = float(split[1])
    #         except Exception:
    #             self.warning('Problem when reading filename %s',f)
    #             continue
    #         temp_list.append( (_T,f) )

    #     #Now sort the numbers
    #     temp_list.sort(key=lambda x: x[0])
    #     return temp_list

    def initialize(self, wngrid):
        """
        Initializes and interpolates the spectral emission density to the current
        stellar temperature and given wavenumber grid

        Parameters
        ----------
        wngrid: :obj:`array`
            Wavenumber grid to interpolate the SED to

        """
        # If temperature outside of range, use blavkbody
        if self.use_blackbody:
            self.warning('Using black body as temperature is outside of Star temeprature range {}'.format(
                self.temperature))
            super().initialize(wngrid)
        else:
            sed = self._base_sed
            self.sed = np.interp(wngrid, self.wngrid, sed)

    @property
    def spectralEmissionDensity(self):
        """
        Spectral emmision density

        Returns
        -------
        sed: :obj:`array`
        """
        return self.sed

    def write(self, output):
        star = super().write(output)
        star.write_string('phoenix_path', self._phoenix_path)
        return star



from .spectrum import BaseSpectrum
import numpy as np
from taurex.util.util import wnwidth_to_wlwidth


class ArraySpectrum(BaseSpectrum):

    """
    Loads an observed spectrum from an array and computes bin
    edges and bin widths. Spectrum shape(nbins, 3-4) with 3-4 columns with
    ordering:
        1. wavelength (um)
        2. spectral data
        3. error
        4. (optional) bin width

    If no bin width is present then they are computed.

    Parameters
    -----------
    filename: string
        Path to observed spectrum file.

    """

    def __init__(self, spectrum):
        super().__init__(self.__class__.__name__)

        self._obs_spectrum = spectrum
        self._bin_widths = None
        self._bin_edges = None

        self._sort_spectrum()
        self._process_spectrum()

        self._wnwidths = wnwidth_to_wlwidth(self.wavelengthGrid,
                                            self._bin_widths)

    def _sort_spectrum(self):
        self._obs_spectrum = \
            self._obs_spectrum[self._obs_spectrum[:, 0].argsort(axis=0)[::-1]]

    def _process_spectrum(self):
        """
        Seperates out the observed data, error, grid and binwidths
        from array. If bin widths are not present then they are
        calculated here
        """
        if self.rawData.shape[1] == 4:
            self._bin_widths = self._obs_spectrum[:, 3]
            obs_wl = self.wavelengthGrid[::-1]
            obs_bw = self._bin_widths[::-1]

            bin_edges = np.zeros(shape=(len(self._bin_widths)*2,))

            bin_edges[0::2] = obs_wl - obs_bw/2
            bin_edges[1::2] = obs_wl + obs_bw/2
            # bin_edges[-1] = obs_wl[-1]-obs_bw[-1]/2.

            self._bin_edges = bin_edges[::-1]
        else:
            self.manual_binning()

    @property
    def rawData(self):
        """Data read from file"""
        return self._obs_spectrum

    @property
    def spectrum(self):
        """The spectrum itself"""
        return self._obs_spectrum[:, 1]

    @property
    def wavelengthGrid(self):
        """Wavelength grid in microns"""
        return self.rawData[:, 0]

    @property
    def wavenumberGrid(self):
        """Wavenumber grid in cm-1"""
        return 10000/self.wavelengthGrid

    @property
    def binEdges(self):
        """ Bin edges"""
        return 10000/self._bin_edges

    @property
    def binWidths(self):
        """bin widths"""
        return self._wnwidths

    @property
    def errorBar(self):
        """ Error bars for the spectrum"""
        return self.rawData[:, 2]

    def manual_binning(self):
        from taurex.util.util import compute_bin_edges
        """
        Performs the calculation of bin edges when none are present
        """
        self._bin_edges, self._bin_widths = \
            compute_bin_edges(self.wavelengthGrid)



"""Functions related to computing emission spectrums"""

import scipy.constants as con
import numpy as np
import ctypes as C
from scipy.stats.mstats_basic import tmean
import numba
import math
from numba import vectorize, float64
from taurex.constants import PI,PLANCK, SPDLIGT, KBOLTZ

@numba.vectorize([float64(float64)],fastmath=True)
def _convert_lamb(lamb):
    return 10000*1e-6/lamb

@numba.vectorize([float64(float64,float64)],fastmath=True)
def _black_body_vec(wl,temp):
    return (PI* (2.0*PLANCK*SPDLIGT**2)/(wl)**5) * (1.0/(np.exp((PLANCK * SPDLIGT) / (wl * KBOLTZ * temp))-1))*1e-6

@numba.njit
def black_body(lamb,temp):
    
    res = np.empty(lamb.shape, dtype=lamb.dtype)
    wl = _convert_lamb(lamb)
#    for i in range(lamb.shape[0]):
#
#        res[i] = (PI* (2.0*PLANCK*SPDLIGT**2)/(wl[i])**5) * (1.0/(math.exp((PLANCK * SPDLIGT) / (wl[i] * KBOLTZ * temp))-1))*1e-6
    
    
    return _black_body_vec(wl,temp)


def black_body_numexpr(lamb, temp):
    import numexpr as ne
    wl = ne.evaluate('10000*1e-6/lamb')
    
    return ne.evaluate('(PI* (2.0*PLANCK*SPDLIGT**2)/(wl)**5) * (1.0/(exp((PLANCK * SPDLIGT) / (wl * KBOLTZ * temp))-1))*1e-6')

def black_body_numpy(lamb, temp):


    h = 6.62606957e-34
    c = 299792458
    k = 1.3806488e-23
    pi= 3.14159265359

    wl = 10000/lamb

    exponent = np.exp((h * c) / (wl*1e-6 * k * temp))
    BB = (pi* (2.0*h*c**2)/(wl*1e-6)**5) * (1.0/(exponent -1))
    return BB * 1e-6
